"""The records module provides the types and protocols for working with
data within your experiment. This module defines only the protocols and
will leave the realization of these protocols to the implementer.
"""

from typing import (
    Protocol,
    Union,
    runtime_checkable,
    Iterable,
    Any)
from adgtk.common import FactoryBlueprint

# ----------------------------------------------------------------------
# Types and Protocols
# ----------------------------------------------------------------------


@runtime_checkable
class PresentableRecord(Protocol):
    """A protocol for a record that is presentable. It ensures that the
    implementation is consistent with the requirements across the
    different objects that will be used to present or process the data.
"""

    def __str__(self) -> str:
        """Generates a string representation

        :return: A string representation of the Record 
        :rtype: str
        """

    def create_copy_of_data(self) -> dict:
        """Creates a copy of the internal data for this record. this
        protects against accidental updating of the data if needed to
        manipulate the values or keys.

        :return: a deep copy of the data
        :rtype: dict
        """

    def copy(self):
        """provides a mapping to __copy__ method. Main benefit is
        readability. no other processing occurs in this method.

        :return: a copy of the record
        :rtype: DataRecord
        """


@runtime_checkable
class SupportsFiltering(Protocol):
    """Provides a consistent definition of a filter for use across modules."""
    blueprint: FactoryBlueprint

    def is_included(self, a: Any) -> bool:
        """Filters a single object. Intended to use as part of filtering
        an iterable object such a list or data store. Implementations
        can opt to use this when they need to define via Factory

        :param record: _description_
        :type record: Any
        :return: _description_
        :rtype: bool
        """


@runtime_checkable
class PresentableGroup(Protocol):
    """One or more PresentableRecords that act together."""
    blueprint: FactoryBlueprint
    records: Iterable[PresentableRecord]
    metadata: dict

    def __len__(self) -> int:
        """The number of records in the group"""

    def __getitem__(self, index: Union[int, slice]):
        """Get an item or a slice

        :param index: The index or slice
        :type index: Union[int, slice]
        """

    def add_record(self, record: PresentableRecord) -> None:
        """Adds a record to the group

        :param record: _description_
        :type record: PresentableRecord
        """

    def create_copy_of_data(self) -> dict:
        """Creates a copy of the internal data for this record. this
        protects against accidental updating of the data if needed to
        manipulate the values or keys.

        :return: a deep copy of the data
        :rtype: dict
        """

    def copy(self):
        """provides a mapping to __copy__ method. Main benefit is
        readability. no other processing occurs in this method.

        :return: a copy of the record group
        :rtype: DataRecordGroup
        """



"""Common, often used patterns. 
"""


from typing import Any, Iterator, Union


# ----------------------------------------------------------------------
#
# ----------------------------------------------------------------------
# py -m pytest -s test/folder/.py


def get_sample_from_iterable(data: Iterator, default: Any = None) -> Any:
    """Gets a sample from an iterable.

    :param data: The iterable data
    :type data: Iterable
    :param default: if unable to get Next, defaults to None
    :type default: Any, optional
    :return: The first item
    :rtype: Any
    """
    return next(data, default)


def get_pair_from_iterable(data: Iterator, default: Any = None) -> tuple:
    """Gets first two entries

    :param data: The iterable data
    :type data: Iterable
    :param default: if unable to get Next, defaults to None
    :type default: Any, optional
    :return: The first and second item
    :rtype: Any
    """
    a = next(data, default)
    b = next(data, default)
    return (a, b)


def string_to_bool(source: Union[str, bool]) -> bool:
    """Converts a string into a boolean. also considers when a bool
    is passed. this keeps the invoking code cleaner.

    :param source: the source string
    :type source: Union[str, bool]
    :raises ValueError: Unable to transform
    :return: the boolean value T/F.
    :rtype: bool
    """
    if isinstance(source, bool):
        return source

    source = source.upper()
    if source == "TRUE":
        return True
    elif source == "FALSE":
        return False
    elif source == "T":
        return True
    elif source == "F":
        return False
    else:
        raise ValueError



"""Welcome to ADGTK (Agentic Data Generation Tool Kit). 

This is a pre-release development version of the project.
"""


__version__ = "0.2.0b1"

__author__ = "Fred Diehl"

__maintainer__ = "Fred Diehl"


# the "main items"
from .utils import get_scenario_logger, create_logger



"""Journal foundation"""

from typing import Protocol, Union


class SupportsReportingOperations(Protocol):
    name: str

    def create_html_and_export(
        self,
        experiment_name: str,
        settings_file_override: Union[str, None] = None,
        header: int = 2,
        base_url: str = "http://127.0.0.1:8000"
    ) -> str:
        """Creates local files and returns HTML that can be used by the
            reports. In addition, exports data, builds images, etc in order
            to save and report.

            :param experiment_name: The experiment name.
            :type experiment_name: str
            :param settings_file_override: the filename of the settings file,
                defaults to None which will use the default file/path.
            :type str, optional
            :param header: The header, defaults to 1
            :type header: int, optional

            :return: HTML that can be used in a report.
            :rtype: str
            """



from .base import (
    InvalidMeasurementConfiguration,
    MeasurementExecution,
    MeasOutputType,
    SupportsMeasDef,
    SupportsMeasSetOps,
    SupportsStopwords,
    MeasurementFeatures,
    Measurement,
    Comparison,
    MeasInputType)
from .engine import MeasurementEngine
from .grouping import MeasurementSet

measurement_register_list = [
    MeasurementEngine,
    MeasurementSet
]



"""The Scenario is the overall wrapper for building an experiment. A
Scenario is also managed by the ScenarioManager, which is responsible
for managing the lifecycle of the Scenario. A proper Scenario should
have at a minimum a method called execute, which will be called by the
ScenarioManager to run the experiment.
"""

from typing import Protocol, runtime_checkable

SCENARIO_GROUP_LABEL = "scenario"


@runtime_checkable
class Scenario(Protocol):
    """The Scenario Protocol"""

    def execute(self, name: str) -> None:
        """Execute the scenario.

        :param name: The name of the experiment (for reporting, etc)
        :type name: str
        """



"""Logging utility functions for configuring and managing loggers."""


import logging
import os
import sys
from typing import Literal, Optional
import warnings
from adgtk.common.defaults import LOG_DIR, SCENARIO_LOGGER_NAME
from .file import clear_folder

def get_scenario_logger() -> logging.Logger:
    """
    Retrieve the logger for the current scenario.

    This utility function ensures that the scenario logger is easily
    accessible and helps maintain clarity in experimentation code.

    If no active scenario is running, it creates a logger using the
    framework directory.

    Returns:
        logging.Logger: The logger instance for the current scenario.
    """
    if SCENARIO_LOGGER_NAME in logging.Logger.manager.loggerDict:
        return logging.getLogger(SCENARIO_LOGGER_NAME)
    
    # not created. using default scenario logger
    print("WARNING: No active scenario found. Using the framework scenario logger")
    logger = create_logger(
        logfile="default.scenario.log",
        logger_name=SCENARIO_LOGGER_NAME,
        log_to_console=False,
        subdir="framework",
        log_propagate=False,
        mode="w"
    )
    logger.info("---- created due to get_scenario_logger w/out a scenario ----")
    return logger

def create_logger(
    logfile: str,
    logger_name: str,
    log_level: int = logging.INFO,
    log_to_console: bool = False,
    subdir: Literal["framework", "runs", "common", "agent"] = "framework",
    experiment_name: Optional[str] = None,
    log_propagate: bool = False,
    mode: Literal["a", "w"] = "a"
) -> logging.Logger:
    """
    Configure and return a logger that writes to a specified log file
    and optionally logs to the console.

    This function removes any existing file handlers for the logger,
    adds a new file handler with the specified log level and format,
    and optionally adds a console handler. It also ensures that the
    necessary directories for logging are created.

    Args:
        logfile (str): Name of the log file to create or append to.
        logger_name (str): Name of the logger instance.
        log_level (int, optional): Logging level (e.g., logging.INFO,
            logging.DEBUG). Defaults to logging.INFO.
        log_to_console (bool, optional): Whether to also log messages
            to the console (stdout). Defaults to False.
        subdir (Literal["framework", "runs", "common"], optional):
            Subdirectory (under 'logs/') to store the log file.
            Defaults to "framework".
        experiment_name (Optional[str], optional): Name of the experiment
            (required if subdir is "runs"). Defaults to None.
        log_propagate (bool, optional): Whether the logger should propagate
            messages to ancestor loggers. Defaults to False.
        mode (Literal["a", "w"], optional): File mode for the log file
            ("a" for append, "w" for write). Defaults to "a".

    Returns:
        logging.Logger: Configured logger instance.

    Raises:
        ValueError: If `experiment_name` is not provided when `subdir`
            is set to "runs".
    """
    if subdir == "agent":
        # special subdir, its runs/exp/agent
        section_dir = os.path.join(LOG_DIR, "runs")
    else:
        section_dir = os.path.join(LOG_DIR, subdir)
    os.makedirs(LOG_DIR, exist_ok=True)
    os.makedirs(section_dir, exist_ok=True)

    log_path = "default"
    if subdir in ["runs", "agent"]:
        if experiment_name is None:
            raise ValueError("Experiment name is required when subdir is 'runs'")
        full_log_dir = os.path.join(section_dir, experiment_name)        
        os.makedirs(full_log_dir, exist_ok=True)
        # modifier for agent
        if subdir == "agent":
            full_log_dir = os.path.join(full_log_dir, "agent")
            os.makedirs(full_log_dir, exist_ok=True)
        log_path = os.path.join(full_log_dir, logfile)
    elif subdir in ["framework", "common"]:
        log_path = os.path.join(section_dir, logfile)

    logger = logging.getLogger(logger_name)

    # Remove all existing handlers for this logger
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)
        handler.close()

    # Add the new handler
    file_handler = logging.FileHandler(log_path, mode=mode)
    formatter = logging.Formatter(
        "%(asctime)s %(levelname)s %(name)s %(message)s")
    file_handler.setFormatter(formatter)
    file_handler.setLevel(log_level)
    logger.addHandler(file_handler)
    logger.setLevel(log_level)
    logger.propagate = log_propagate

    if log_to_console:
        if not any(
                isinstance(h, logging.StreamHandler) for h in logger.handlers):
            stream_handler = logging.StreamHandler(sys.stdout)
            stream_handler.setFormatter(formatter)
            stream_handler.setLevel(log_level)
            logger.addHandler(stream_handler)

    return logger


def set_logfile(
    logfile: str,
    logger_name: str,
    log_level: int = logging.INFO,
    log_to_console: bool = False,
    subdir: Literal["framework", "runs", "common", "agent"] = "framework",
    experiment_name: Optional[str] = None,
    log_propagate: bool = True,
    mode: Literal["a", "w"] = "a"
) -> logging.Logger:
    """
    [DEPRECATED] Configure and return a logger that writes to a specified
    log file and optionally logs to the console.

    This function is deprecated as of version 0.2. Use `create_logger`
    instead. It retains the same functionality as `create_logger`.

    Args:
        logfile (str): Name of the log file to create or append to.
        logger_name (str): Name of the logger instance.
        log_level (int, optional): Logging level (e.g., logging.INFO,
            logging.DEBUG). Defaults to logging.INFO.
        log_to_console (bool, optional): Whether to also log messages
            to the console (stdout). Defaults to False.
        subdir (Literal["framework", "runs", "common"], optional):
            Subdirectory (under 'logs/') to store the log file.
            Defaults to "framework".
        experiment_name (Optional[str], optional): Name of the experiment
            (required if subdir is "runs"). Defaults to None.
        log_propagate (bool, optional): Whether the logger should propagate
            messages to ancestor loggers. Defaults to True.
        mode (Literal["a", "w"], optional): File mode for the log file
            ("a" for append, "w" for write). Defaults to "a".

    Returns:
        logging.Logger: Configured logger instance.

    Raises:
        ValueError: If `experiment_name` is not provided when `subdir`
            is set to "runs".

    Warnings:
        DeprecationWarning: This function is deprecated. Use `create_logger`
        instead.
    """
    warnings.warn(
        "set_logfile is deprecated. Use create_logger instead.",
        DeprecationWarning,
        stacklevel=2
    )
    return create_logger(
        logfile=logfile,
        logger_name=logger_name,
        log_level=log_level,
        log_to_console=log_to_console,
        subdir=subdir,
        experiment_name=experiment_name,
        log_propagate=log_propagate,
        mode=mode
    )


# ----------------------------------------------------------------------
# Color Logger
# ----------------------------------------------------------------------

role_types = Literal["human", "ai", "system", "tool", "error", "default"]

class RoleColorFormatter(logging.Formatter):
    """Formats based on role. Sets the color."""
    ROLE_COLORS = {
        'human': '\033[94m',    # Blue
        'ai': '\033[92m',       # Green
        'system': '\033[95m',   # Magenta
        'tool': '\033[96m',     # Cyan
        'error': '\033[91m',    # Red
        'default': '\033[0m',   # Reset
    }

    def format(self, record):
        role = getattr(record, 'role', 'default')
        color = self.ROLE_COLORS.get(role, self.ROLE_COLORS['default'])
        line = "-" * 30
        bold_role = f"\033[1m{role.upper()}\033[0m"        

        # Split the message into lines
        message_lines = str(record.msg).splitlines()
        if message_lines:
            # Format the first line with the role and color
            message_lines[0] = f"{line} {color}{bold_role} {line}\n{message_lines[0]}\033[0m"
        # Join the lines back together for display
        formatted_message = "\n".join(message_lines)

        # Use the formatted message for output, without modifying record.msg
        return formatted_message



def clear_llm_logs(experiment_name:str) -> None:
    """Clears the LLM dedicated log folder for an experiment

    Args:
        experiment_name (str): The experiment to clear
    """
    section_dir = os.path.join(LOG_DIR, "runs")
    full_log_dir = os.path.join(section_dir, experiment_name)
    llm_dir = os.path.join(full_log_dir, "llm")
    try:
        clear_folder(llm_dir)
    except FileNotFoundError:
        pass

def clear_agent_logs(experiment_name:str) -> None:
    """Clears the agent dedicated log folder for an experiment

    Args:
        experiment_name (str): The experiment to clear
    """
    section_dir = os.path.join(LOG_DIR, "runs")
    full_log_dir = os.path.join(section_dir, experiment_name)
    agent_dir = os.path.join(full_log_dir, "agent")
    try:
        clear_folder(agent_dir)
    except FileNotFoundError:
        pass


def create_llm_logger(
    logfile: str,
    logger_name: str,
    experiment_name: str,    
    log_level: int = logging.INFO,
    log_to_console: bool = True,
    log_propagate: bool = False,
    mode: Literal["a", "w"] = "a"
) -> logging.Logger:
    """
    Configure and return a logger with role-based color formatting
    that writes to a specified log file and optionally logs to the console.

    Args:
        logfile (str): Name of the log file to create or append to.
        logger_name (str): Name of the logger instance.
        log_level (int, optional): Logging level (e.g., logging.INFO,
            logging.DEBUG). Defaults to logging.INFO.
        log_to_console (bool, optional): Whether to also log messages
            to the console (stdout). Defaults to True.
        subdir (Literal["framework", "runs", "common"], optional):
            Subdirectory (under 'logs/') to store the log file.
            Defaults to "framework".
        experiment_name (Optional[str], optional): Name of the experiment
            (required if subdir is "runs"). Defaults to None.
        log_propagate (bool, optional): Whether the logger should propagate
            messages to ancestor loggers. Defaults to False.
        mode (Literal["a", "w"], optional): File mode for the log file
            ("a" for append, "w" for write). Defaults to "a".

    Returns:
        logging.Logger: Configured logger instance with role-based color formatting.
    """
    section_dir = os.path.join(LOG_DIR, "runs")
    os.makedirs(LOG_DIR, exist_ok=True)
    os.makedirs(section_dir, exist_ok=True)
    
    full_log_dir = os.path.join(section_dir, experiment_name)
    os.makedirs(full_log_dir, exist_ok=True)
    llm_dir = os.path.join(full_log_dir, "llm")
    os.makedirs(llm_dir, exist_ok=True)
    log_path = os.path.join(llm_dir, logfile)
    logger = logging.getLogger(logger_name)

    # Remove all existing handlers for this logger
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)
        handler.close()

    msg_formatter = RoleColorFormatter("%(message)s")
    file_handler = logging.FileHandler(log_path, mode=mode)
    file_handler.setFormatter(msg_formatter)
    file_handler.setLevel(log_level)
    logger.addHandler(file_handler)
    logger.setLevel(log_level)
    logger.propagate = log_propagate

    if log_to_console:
        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setFormatter(msg_formatter)
        stream_handler.setLevel(log_level)
        logger.addHandler(stream_handler)

    return logger


"""Tracks various performance statistics and writes to disk"""

import logging
import os
import csv
from typing import Union, Any
from adgtk.common import FolderManager
from adgtk.journals import ExperimentJournal
from .base import MetricTracker


class PerformanceTracker():
    """Used to track and report on running performance. Examples include
    success rate, average time to complete a task, etc. The design is to
    be as flexible as possible to support any future user needs.
    """


    def __init__(
        self,
        experiment_name:str,
        component:str,
        last_only:bool=True,
        journal: Union[ExperimentJournal, None] = None
    ):
        """Create a new PerformanceTracker object.

        :param experiment_name: The name of the experiment
        :type experiment_name: str
        :param component: The name of the component recording the data
        :type component: str
        :param journal: The experiment journal. defaults to None
        :type journal: ExperimentJournal
        :param last_only: return only the last value, defaults to True
        :type last_only: bool, optional      
        """
        
        
        super().__init__()
        self.last_only = last_only
        self.metric_tracker = MetricTracker()
        self.folders = FolderManager(experiment_name)
        self.component = component
        self.journal =journal

    def register_statistic(self, label:str):
        """Registers a statistic. Enables tracking of values using this
        label. This allows for fetch by the label as well as the ability
        to save this data by label to disk.

        :param label: The label for the statistic
        :type label: str
        """
        self.metric_tracker.register_metric(label=label)

    def add_data(self, label:str, value:float|int):
        """Add data to the performance tracker for a given label. If the
        label is not found the method will attempt to register the label
        then add the data.

        :param label: the label for the statistic
        :type label: str
        :param value: The value to record
        :type value: float | int
        """
        try:
            self.metric_tracker.add_data(label=label, value=value)
        except KeyError:
            # likely tried to add data to an unregistered label.
            # One last try. let raise occur if happens again.
            self.metric_tracker.register_metric(label=label)
            self.metric_tracker.add_data(label=label, value=value)

        
    def save_data(self):
        """Saves the data to disk using the pre-defined folder structure
        """
        # prepare data
        labels = self.metric_tracker.metric_labels()
        out_data = {}
        for label in labels:
            # always save all data
            data = self.metric_tracker.get_all_data(label)                
            # now set the data, if exists
            if len(data) > 0:
                out_data[label] = data
            else:
                msg = f"{self.component} metric tracker had no data recorded "\
                      f"for {label}"
                logging.warning(msg)                
                out_data[label] = []

        # write to disk        
        for key, data in out_data.items():
            filename = os.path.join(
                        self.folders.performance,
                        f"{self.component}.{key}.csv")

            with open(filename, "w", newline="") as outfile:
                writer = csv.writer(outfile)
                writer.writerow(data)

                if self.journal is not None:
                    # Log the write to the report
                    self.journal.log_data_write(
                        description=f"{self.component}.{key} metric data",
                        file_w_path=filename,
                        component=f"{self.component}.{key}",
                        entry_type="metrics")



"""component.py is the new factory design introduced with the v0.2.0
re-write. The goal of this re-write is to greatly simplify the amount of
boilerplate code needed by grounding the design with pydantic + Protocol

This is intended to be non-persistent so each run will need to handle
their own assembly.

Goals
=====
A dynamic factory for both internal and user defined objects.

Testing
=======
py -m pytest -s test/factory/test_component_factory.py

Notes
=====
1. MVP design.
2. Goal remains the ability to intermix both framework provided with the
   user provided entries.
3. No persistent storage. The factory is loaded via the CLI by the
   bootstrap.py file. The design has the user update their bootstrap
   file.
4. When to log. when a raise is about communication, no log, else log

Roadmap
=======
1. Consider internal_only objects that are hidden from reports as well
   as denied the ability to register if outside of this project. i.e. a
   user cannot register additional internal-only objects group.
2. Consider an experiment definition override to another bootstrap file.
3. on the report, between sections increase the line length to match.

Defects
=======
1.
"""
import os
import sys
# before importing others
# ----------------------------------------------------------------------
# Start of path verification
# ----------------------------------------------------------------------
path = os.getcwd()
bootstrap_file = os.path.join(path, "bootstrap.py")
if not os.path.exists(bootstrap_file):
    print("ERROR: Unable to locate the bootstrap.py. Please check your path.")
    sys.exit(1)
# ----------------------------------------------------------------------
# End of path verification
# ----------------------------------------------------------------------

from adgtk.factory.structure import (
    BlueprintQuestion,
    SupportsFactory,
    FactoryEntry,
    FactoryOrder
)
from pydantic import ValidationError
from typing import (
    cast,
    Callable,
    Optional,
    Type,
    Union)
import secrets
import inspect
import copy

# setup logfile for this and sub-modules
from adgtk.utils import create_logger

# Set up module-specific logger
_logger = create_logger(
    "adgtk.factory.log",
    logger_name=__name__,
    subdir="framework"
)


# ----------------------------------------------------------------------
# Globals
# ----------------------------------------------------------------------

_inventory: dict[str, FactoryEntry] = {}  # all constructors
_groups: list[str] = []                   # only grows, consider dict[str,int]


# ----------------------------------------------------------------------
# Decorator
# ----------------------------------------------------------------------
def register_to_factory(cls):
    """A decorator for registration to the factory"""
    if issubclass(cls, SupportsFactory):
        register(
            item=cls,
            group=cls.group,
            tags=cls.tags,
            factory_id=cls.factory_id,
            summary=cls.summary,
            interview_blueprint=cls.interview_blueprint,
            factory_can_init=cls.factory_can_init
        )
        return cls

    raise ValueError(
        "This decorator supports children of SupportsFactory class")

# ----------------------------------------------------------------------
# Public
# ----------------------------------------------------------------------


def register(
    item: Union[Callable, SupportsFactory],
    group: Optional[str] = None,
    tags: Optional[list] = None,
    factory_id: Optional[str] = None,
    summary: str = "No summary recorded",
    interview_blueprint: Optional[list[BlueprintQuestion]] = None,
    factory_can_init: Optional[bool] = None
) -> str:
    """Registers an item into the factory.

    :param item: The callable item
    :type item: Union[Callable, SupportsFactory]
    :param group: the group name if not SupportsFactory. Required if
        item is not SupportsFactory. If SupportsFactory it overrides
        the item value for the group. defaults to None
    :type group: Optional[str], optional
    :param tags: tags for searching the factory, defaults to None
    :type tags: Optional[list], optional
    :param factory_id: The id for the factory, defaults to None
    :type factory_id: Optional[str], optional
    :param summary: The summary for listings, defaults to "No summary recorded"
    :type summary: str, optional
    :param interview_blueprint: The questions to ask for experiment
        definitions, defaults to None
    :type interview_blueprint: Optional[list[BlueprintQuestion]], optional
    :raises ValueError: Invalid request, failure to process, etc    
    :raises IndexError: Entry already exists
    :return: The factory id now entered into the factory
    :rtype: str
    """
    global _inventory, _groups

    # set the factory_can_init
    if factory_can_init is None:
        # If not set then check, if not SupportsFactory then false
        factory_can_init = False
        if inspect.isclass(item) and issubclass(item, SupportsFactory):
            factory_can_init = item.factory_can_init

    if not callable(item):
        raise ValueError("Invalid item. It must be callable.")

    # shorter than UUID.
    factory_id = factory_id or getattr(
        item, "factory_id", f"tmp.{secrets.token_hex(4)}")

    if factory_id is None:
        msg = "Failed to create factory_id"
        _logger.error(msg)
        raise ValueError(msg)

    # verify formatting
    if isinstance(factory_id, int):
        msg = "Invalid factory_id. Must not be able to convert to int"
        _logger.info(msg)
        raise ValueError(msg)
    try:
        _ = int(factory_id)
    except ValueError:
        pass
    else:
        msg = (f"Invalid factory_id {factory_id}. Must not be able "
               "to convert to int")
        _logger.info(msg)
        raise ValueError(msg)

        pass
    if factory_id in _inventory:
        raise IndexError(f"factory_id: {factory_id} already exists")

    if (inspect.isclass(item) and issubclass(item, SupportsFactory)):
        if summary == "No summary recorded":
            summary = item.summary

        group = group or item.group
        tags = (tags or []) + item.tags
        interview_blueprint = item.interview_blueprint
    else:
        if group is None:
            raise ValueError("Group required if not using SupportsFactory.")
        tags = tags or []
        if interview_blueprint is None:
            interview_blueprint = []

    entry = FactoryEntry(
        factory_id=factory_id,
        factory_can_init=factory_can_init,
        creator=item,
        group=group,
        tags=tags,
        interview_blueprint=interview_blueprint,
        summary=summary
    )

    if group not in _groups:
        _groups.append(group)
    if entry.factory_id is None:
        raise ValueError("entry construction error. missing factory_id")

    _inventory[entry.factory_id] = entry
    msg = f"Registered factory_id: {factory_id}"
    _logger.info(msg)
    return entry.factory_id


def create_using_order(order: FactoryOrder) -> SupportsFactory:
    """Creates an object using a FactoryOrder. This provides an easier
    interface for Scenario loading, etc. This does not work for fetching
    a callable.

    :param order: The order to process
    :type order: FactoryOrder
    :return: The resulting object with the values set in the init_args
    :rtype: T
    """
    if not isinstance(order, FactoryOrder):
        try:
            order = FactoryOrder(**order)
        except ValidationError as e:
            _logger.error("Invalid order submitted")
            raise

    if order.init_args is None:
        return create(factory_id=order.factory_id)

    init_args = order.init_args
    created = create(factory_id=order.factory_id, **init_args)
    return created


def create(factory_id: str, **kwargs) -> SupportsFactory:
    """Creates an instance of the item

    :param factory_id: The id to create
    :type factory_id: str
    :raises KeyError: Unable to find in the factory
    :return: An instance of the factory item
    :rtype: T
    """
    global _inventory

    if factory_id not in _inventory:
        msg = f"{factory_id} not in factory"
        _logger.error(msg)
        raise KeyError(msg)

    item = _inventory[factory_id]
    creator = item.creator
    if item.factory_can_init:
        created = creator(**kwargs)
        return created

    msg = (f"Attempted to init factory_id: {item.factory_id} that does "
           "not support init.")
    _logger.error(msg)
    raise ValueError(msg)


def get_interview(factory_id: str) -> list[BlueprintQuestion]:
    """Provides an interview for a factory item based on the
    data provided at registration

    :param factory_id: The id to obtain the interview for
    :type factory_id: str
    :raises KeyError: Unable to find the id in the factory
    :return: The interview
    :rtype: list[BlueprintQuestion]
    """
    if factory_id not in _inventory:
        msg = f"{factory_id} not in factory"
        _logger.error(msg)
        raise KeyError(msg)

    item = _inventory[factory_id]
    return item.interview_blueprint


def get_callable(factory_id: str) -> Callable:
    """Returns a callable. It does not attempt to initialize.

    :param factory_id: The id of the item to fetch
    :type factory_id: str
    :raises KeyError: Unable to find the factory id
    :return: a callable object, ex function. does not invoke
    :rtype: Callable
    """
    global _inventory

    if factory_id not in _inventory:
        msg = f"{factory_id} not in factory"
        _logger.error(msg)
        raise KeyError(msg)

    entry = _inventory[factory_id]
    return entry.creator


def remove(factory_id: str) -> None:
    """Removes an item from the factory

    :param factory_id: The id of the item to remove
    :type factory_id: str
    :raises KeyError: Unable to locate id
    """

    global _inventory

    if not factory_id in _inventory.keys():
        msg = f"{factory_id} not in factory"
        _logger.error(msg)
        raise KeyError(msg)

    # TODO: in the future, consider removing from _groups
    del _inventory[factory_id]


def list_entries(
    tags: Optional[Union[str, list[str]]] = None,
    group: Optional[str] = None
) -> list:
    """Lists the entries within the factory

    :param tags: Filter on tag(s), defaults to None
    :type tags: Optional[Union[str, list[str]]], optional
    :param group: Filter on a group, defaults to None
    :type group: Optional[str], optional
    :raises ValueError: Corruption of the inventory
    :return: A list of entries that match the search
    :rtype: list
    """
    global _inventory

    found: list[FactoryEntry] = []
    entry: FactoryEntry

    # cleanup input
    if isinstance(tags, str) and len(tags) > 0:
        tags = [tags]

    # this should always return true but just in case:
    if isinstance(tags, list):
        # cleanup any white-space
        tags = [x.strip() for x in tags]

        # and if all that is left is an empty string
        for tag in tags:
            if tag == "":
                tags.remove("")

    for _, entry in sorted(_inventory.items(), key=lambda x: x[0]):
        if not isinstance(entry, FactoryEntry):
            try:
                entry = FactoryEntry(**entry)
            except ValidationError:
                msg = "The Inventory appears to be corrupted"
                _logger.error(msg)
                raise ValueError(msg)

        if tags is None:
            if group is None:
                found.append(entry)
            elif group.lower() == entry.group.lower():
                found.append(entry)
        elif all(entry_tag in entry.tags for entry_tag in tags):
            if group is None:
                found.append(entry)
            elif group == entry.group:
                found.append(entry)
    return found


def report(
    tags: Optional[Union[str, list[str]]] = None,
    group: Optional[str] = None
) -> None:
    """Generates a report and prints to console of all the files that
    are curently in the inventory.

    :param tags: The tags to filter for, defaults to None
    :type tags: Optional[Union[str, list]], optional
    """
    global _groups

    title = "Factory report"
    if group is not None:
        title += f" - group={group}"
    if tags is not None:
        if isinstance(tags, list):
            tag_str = " ".join(tags)
            title += f", tags={tag_str}"
        else:
            title += f", tag={tags}"

    longest = 0
    all_entries = ""
    first = True
    if group is not None:
        search_groups = [group]
    else:
        _groups.sort()
        search_groups = _groups

    for group in search_groups:
        if first:
            all_entries += f"{group.upper()}\n"
            first = False
        else:
            all_entries += "-"*79
            all_entries += f"\n{group.upper()}\n"

        entries = list_entries(tags=tags, group=group)
        for entry in entries:
            entry_str = f"  - {entry.factory_id:<15} | "
            if entry.factory_can_init:
                entry_str += " Y  |  "
            else:
                entry_str += " N  |  "
            entry_str += f"{entry.summary:50} |"

            tags_str = ""
            if entry.tags is not None:
                tags_str = " ".join(entry.tags)
            entry_str += tags_str

            if len(entry_str) > longest:
                longest = len(entry_str)

            all_entries += f"{entry_str}\n"

    # Setup title/'banner', the spaced out title for the columns
    name = "Factory ID"
    summary = "Summary"
    banner = f"    {name:<15} | init | {summary:<50} | Tags"
    if len(title) > longest:
        longest = len(title)
    if len(banner) > longest:
        longest = len(banner)

    if longest > len(title):
        # now center the title
        spaces = int((longest-len(title))/2)
        space_str = " "*spaces
        title = f"\n{space_str}{title}"
    # and finally put everything together and print
    line = "="*longest
    small_line = "-"*longest
    title += f"\n{line}\n{banner}\n{small_line}\n{all_entries}"
    print(title)


def entry_exists(factory_id: str) -> bool:

    return factory_id in _inventory


def group_exists(group: str) -> bool:
    return group in _groups


def get_group_names() -> list:
    return copy.deepcopy(_groups)



"""The dataset module is focused on providing a consistent interface for
registering, loading, and managing datasets.  It accomplishes this by
inheriting the JsonFileTracker and using the file utilities provides an
easy to use interface for registering a file and loading the file by the
id.

Roadmap
=======
1. Improved reporting by use
"""

import os
from typing import Literal, Optional, Union
from adgtk.data.structure import FileEncodingTypes, PurposeTypes
from adgtk.data.tracking import JsonFileTracker
import adgtk.tracking.journal as exp_journal
from adgtk.data.utils import load_data_from_file, ReturnDataTypes
from adgtk.utils import create_logger


FILENAME = "datasets.json"


class DatasetManager(JsonFileTracker):

    def __init__(
        self,
        name:str = "dataset.manager",
        folder: str = ".tracking",
        inventory_file: Optional[str] = None
    ) -> None:
        os.makedirs(folder, exist_ok=True)
        file_name = inventory_file or FILENAME
        file_w_path = os.path.join(folder, file_name)
        self.logger = create_logger(
            logfile=f"{name}.log",
            logger_name=name,
            subdir="common"
        )
        super().__init__(
            label=name,
            inventory_file=file_w_path,
            logger=self.logger
        )

    def register(
        self,
        source_file: str,
        encoding: FileEncodingTypes,
        metadata_file: Optional[str] = None,
        tags: Optional[Union[str, list[str]]] = None,
        id: Optional[str] = None,
        use: Literal["test", "train", "validate", "other"] = "other",
        purpose:PurposeTypes= "other"
    ) -> str:
        """Registers a file definition.

        :param source_file: The name of the file w/path
        :type source_file: str
        :param encoding: The encoding of the file
        :type encoding: FileEncodingTypes
        :param tags: The tags for this file, defaults to None
        :type tags: Optional[list[str]], optional
        :param id: The requested ID, defaults to None
        :type id: Optional[str], optional
        :param purpose: The purpose of the file
        :type purpose: PurposeTypes
        :raises FileNotFoundError: File is not found
        :raises IndexError: The id already exists
        :return: The id of the file saved
        :rtype: str
        """
        full_path = os.path.abspath(source_file)
        if not os.path.exists(full_path):
            raise FileNotFoundError("unable to find file: %s", full_path)

        if tags is None:
            tags = []
        elif isinstance(tags, str):
            tags = [tags]
        tags.append(use)
        exp_journal.add_file(filename=full_path, purpose=purpose)

        return self.register_file(
            source_file=full_path,
            encoding=encoding,
            metadata_file=metadata_file,
            tags=tags,
            id=id
        )


    def load_file(self, id: str) -> ReturnDataTypes:
        """Pulls the file definition from the file tracker and then uses
        the load_data_from_file utility function to load the data.

        :param id: the ID in the file_tracker.
        :type id: str
        :return: The loaded data
        :rtype: ReturnDataTypes
        """
        file_def = self.get_file_definition(id)
        return load_data_from_file(file_def=file_def)



"""journal.py is intended as the common journal for the experiment."""

import json
import os
from typing import Optional
import datetime
from adgtk.data.structure import FileEntry, PurposeTypes
from adgtk.tracking.structure import CommentModel


# ----------------------------------------------------------------------
# Globals
# ----------------------------------------------------------------------

# data
_files_written: list[FileEntry] = []
_comments:list[CommentModel] = []

# ----------------------------------------------------------------------
# Helper functions
# ----------------------------------------------------------------------

def _get_components() ->set[str]:
    """Obtains a listing (via a set) of the components recorded.

    :return: A set of the components within the comments
    :rtype: set[str]
    """
    found = []
    for entry in _comments:
        found.append(entry.component)

    return set(found)


# ----------------------------------------------------------------------
# Public Functions
# ----------------------------------------------------------------------

def reset() -> None:
    """Resets the journal tracking. primary purpose would be when doing
    batch processing and you have multiple scenarios.
    """
    global _files_written
    global _comments
    _files_written = []
    _comments = []


def add_file(filename:str, purpose:PurposeTypes) -> None:
    """Writes a file save to the journal

    Args:
        filename (str): _description_
        purpose (PurposeTypes): _description_
    """
    file = FileEntry(filename=filename, purpose=purpose)
    if file not in _files_written:
        _files_written.append(file)

def add_comment(
    comment:str,
    use_timestamp:bool=True,
    component:Optional[str]=None
) -> None:
    """Adds a comment to the journal

    :param comment: The comment to add
    :type comment: str
    :param use_timestamp: include timestamp in the statement, defaults to True
    :type use_timestamp: bool, optional
    :param component: a component tag, defaults to None
    :type component: Optional[str], optional
    """
    now = None
    if use_timestamp:
        now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    if component is None:
        entry = CommentModel(comment=comment, timestamp=now)
    else:
        entry = CommentModel(
            comment=comment, timestamp=now, component=component)

    if entry not in _comments:
        _comments.append(entry)


def save_journal(path:str, filename:str="journal.json") -> None:
    """Saves the journal to disk as a json file.

    :param path: the path for where to write
    :type path: str
    :param filename: The filename to use, defaults to "journal.json"
    :type filename: str, optional
    """

    file_w_path = os.path.join(path, filename)
    comments_as_dicts = [comment.model_dump() for comment in _comments]
    files_as_dicts = [file.model_dump() for file in _files_written]
    data = {
        "comments": comments_as_dicts,
        "files": files_as_dicts
    }
    # Now write to JSON
    with open(file=file_w_path, mode='w', encoding="utf-8") as outfile:
        json.dump(files_as_dicts, outfile, indent=2)

def generate_report(
    path:str,
    experiment_name:str,
    filename:str="report.html"
) -> None:
    """Generates a report. This is more a human readable version. The
    save_journal and its json is the truly the target for an Agent to
    use for review of the experiment.

    :param path: The path on where to write the report
    :type path: str
    :param The name of the experiment: _description_
    :type experiment_name: str
    :param filename: The filename to use, defaults to "report.html"
    :type filename: str, optional
    """
    components = _get_components()


    # old code from previous design    

    # template = env.get_template("report.jinja")
    # try:
    #     output = template.render(
    #         date_ran=date_ran,
    #         experiment_name=experiment_name,
    #         comments=self._comments,
    #         tools=tools,
    #         measurement_section=measurement_html,
    #         scenario_def=self._scenario_def,
    #         data_section=data_html)
    #     with open(
    #             file=report_filename,
    #             encoding="utf-8",
    #             mode="w") as outfile:
    #         outfile.write(output)
    # except jinja2.exceptions.TemplateSyntaxError as e:
    #     logging.error("Syntax error with report.jinja")



"""Internal tracking. Provides useful data structures.
"""

import copy
import csv
import os
from typing import Iterable, Union
import numpy as np
from adgtk.data.structure import PurposeTypes
import adgtk.tracking.journal as exp_journal
from adgtk.utils import get_scenario_logger
from .structure import ExperimentRunFolders
# ----------------------------------------------------------------------
# Constants
# ----------------------------------------------------------------------
DEBUG_TO_CONSOLE = False

# ----------------------------------------------------------------------
# Tracking of data
# ----------------------------------------------------------------------

class MetricTracker():
    """Used for tracking metrics."""

    def __init__(
        self,
        name:str="experiment",
        purpose:PurposeTypes="other"
    ):
        self.name = name    # for saving to disk
        self.purpose: PurposeTypes = purpose
        self.metrics: dict[str, list] = {}
        self.metadata: dict[str, dict] = {}
        self.logger = get_scenario_logger()

    def register_metric(
        self,
        label: str,
        metadata: Union[dict, None] = None
    ) -> bool:
        """Registers a metric.

        :param label: The label of the metric
        :type label: str
        :param metadata: _description_
        :type metadata: Union[dict, None]
        :return: T: created, F: did not create
        :rtype: bool
        """

        if metadata is not None:
            if label not in self.metadata:
                self.metadata[label] = metadata
        else:
            if label not in self.metadata:
                self.metadata[label] = {}

        if label not in self.metrics:
            self.metrics[label] = []
            return True
        return False

    def add_raw_data(self, label: str, values: Iterable) -> None:
        """Adds data as-is by iterating through and adding one by one.

        :param label: The label of the metric
        :type label: str
        :param values: the data to add
        :type values: Iterable
        """
        for data in values:
            self.add_data(label=label, value=data)

    def add_data(self, label: str, value: Union[int, float]) -> None:
        """Adds data

        :param label: The label of the metric
        :type label: str
        :raises KeyError: Label is not found
        :param value: the data to add
        :type Union[int, float]
        """
        if label not in self.metrics:
            self.metrics[label] = []

        self.metrics[label].append(value)

        if DEBUG_TO_CONSOLE:
            print(f"MetricTracker adding {value} to {label}")
            print(f"Updated Metrics: {self.metrics[label]}")

    def metric_exists(self, label: str) -> bool:
        """Does a metric exist?

        :param label: The label of the metric
        :type label: str
        :return: T: exists, F: does not
        :rtype: bool
        """
        if label not in self.metrics:
            return False

        return True

    def remove_metric(self, label: str) -> None:
        """Removes a metric from being tracked

        :param label: The label of the metric to remove
        :type label: str
        """
        if label in self.metrics:
            del self.metrics[label]

        if label in self.metadata:
            del self.metadata[label]

    def metric_labels(self) -> list:
        """Gets a list of metrics currently tracking

        :return: a list of the labels
        :rtype: list
        """
        return list(self.metrics.keys())

    def get_latest_value(self, label: str) -> float:
        """Gets the latest value from a label

        :param label: The label of the metric to get latest value
        :type label: str
        :raises KeyError: Invalid Metric
        :return: the latest value
        :rtype: float
        """
        if label not in self.metrics:
            msg = f"Requested invalid label: {label}"
            self.logger.error(msg)
            raise KeyError("Invalid metric")
        elif len(self.metrics[label]) == 0:
            return 0
        else:
            return self.metrics[label][-1]
        
    def get_latest_distribution(self, label: str) -> np.ndarray:
        """Gets the latest distribution from a label

        :param label: The label of the metric to get latest value
        :type label: str
        :raises KeyError: Invalid Metric
        :return: the latest value
        :rtype: np.ndarray
        """
        if label not in self.metrics:
            msg = f"Requested invalid label: {label}"
            self.logger.error(msg)
            raise KeyError("Invalid metric")
        elif len(self.metrics[label]) == 0:
            return np.ndarray([])
        else:
            return self.metrics[label][-1]

    def get_average(self, label: str) -> float:
        """Returns the average of all stored values for the label

        :param label: The label of the metric to get avg of
        :type label: str
        :raises KeyError: Invalid Metric
        :return: the average value
        :rtype: float
        """
        if label not in self.metrics:
            msg = f"Requested invalid label: {label}" 
            if DEBUG_TO_CONSOLE:
                print(f"METRIC_TRACKER_DATA: {self.metrics}")
                print(msg)
            self.logger.error(msg)
            raise KeyError("Invalid metric")
        elif len(self.metrics[label]) == 0:
            return 0
        else:
            return sum(self.metrics[label]) / len(self.metrics[label])

    def get_sum(self, label: str) -> float:
        """Returns the sum of all stored values for the label

        :param label: The label of the metric to get sum of
        :type label: str
        :raises KeyError: Invalid Metric
        :return: the sum of all stored values
        :rtype: float
        """
        if label not in self.metrics:
            msg = f"Requested invalid label: {label}"
            self.logger.error(msg)
            raise KeyError("Invalid metric")
        elif len(self.metrics[label]) == 0:
            return 0
        else:
            return sum(self.metrics[label])

    def clear_metric(self, label: str) -> None:
        """Clears the values of a  metric

        :param label: The label of the metric to clear data from
        :type label: str
        """
        self.metrics[label] = []

    def clear_results(self) -> None:
        """Clears all prior measurement results
        """
        for key in self.metrics.keys():
            self.metrics[key] = []

    def reset(self) -> None:
        """Deletes all data and labels and resets to no metrics tracked.
        """
        self.metrics = {}

    def measurement_count(self, label: str) -> int:
        """Returns the count of observations for a metric


        :param label: The label of the metric to get count of
        :type label: str
        :raises KeyError: Invalid Metric
        :return: the count of all entries
        :rtype: int
        """
        if label not in self.metrics:
            raise KeyError("Invalid metric")

        return len(self.metrics[label])

    def get_all_data(self, label: str) -> list:
        """Gets all data for a metric.

        :param label: The label of the metric to get data from
        :type label: str
        :raises KeyError: Invalid Metric
        :return: the data as a list
        :rtype: list
        """
        if self.metric_exists(label):
            return copy.deepcopy(self.metrics[label])

        msg = f"Requested invalid label: {label}"
        self.logger.error(msg)
        raise KeyError("Invalid metric")

    def get_metadata(self, label: str) -> dict:
        if label in self.metadata:
            return copy.deepcopy(self.metadata[label])

        msg = f"Requested invalid metadata for label: {label}"
        self.logger.error(msg)
        return {}

    def save_data(self, folders: ExperimentRunFolders) -> None:
        """Saves the data to disk using the pre-defined folder structure

        Args:
            folders (ExperimentRunFolders): The results folders
        """
        # prepare data
        labels = self.metric_labels()
        out_data = {}
        for label in labels:
            # always save all data
            data = self.get_all_data(label)                
            # now set the data, if exists
            if len(data) > 0:
                out_data[label] = data
            else:
                msg = f"{self.name} metric tracker had no data recorded "\
                      f"for {label}"
                self.logger.warning(msg)                
                out_data[label] = []

        # write to disk        
        for key, data in out_data.items():
            filename = os.path.join(
                        folders.metrics,
                        f"{self.name}.{key}.csv")

            with open(filename, "w", newline="") as outfile:
                writer = csv.writer(outfile)
                writer.writerow(data)

            self.logger.info(
                f"Saved {self.name}.{key} metric data to {filename}")

            exp_journal.add_file(filename=filename, purpose=self.purpose)

    def export_last_val_to_dict(self) -> dict:
        """Obtains the latest measurement as a dictionary.

        Returns:
            dict: the last value recorded by metric label
        """        
        # prepare data
        labels = self.metric_labels()
        out_data = {}        
        for label in labels:
            out_data[label] = self.get_latest_value(label)
        return out_data



"""Functions for the factory"""

# ----------------------------------------------------------------------
# Demo functions
# ----------------------------------------------------------------------

def point_five_function(s1:str, s2:str) -> float:
    """An extremely basic function for validation of factory. Primary
    used for demonstration and testing. Limited to no utility.
    
    Returns 1 if both strings are the same, else .5

    :param s1: first string
    :type s1: str
    :param s2: second string
    :type s2: str
    :return: either 1 or .5
    :rtype: float
    """
    if s1 == s2:
        return 1
    return 0.5

def point_two_function(s1:str, s2:str) -> float:
    """An extremely basic function for validation of factory. Primary
    used for demonstration and testing. Limited to no utility.
    
    Returns 1 if both strings are the same, else .2

    :param s1: first string
    :type s1: str
    :param s2: second string
    :type s2: str
    :return: either 1 or .2
    :rtype: float
    """
    if s1 == s2:
        return 1
    return 0.2


# """Common module
# """
# from .exceptions import (
#     DuplicateFactoryRegistration,
#     InvalidConfigException,
#     InvalidScenarioState,
#     InvalidBlueprint,
#     InsufficientData)
# from .structure import (
#     ToolDefinition,
#     ExperimentDefinition,
#     ComponentDef,
#     ArgumentType,
#     ArgumentSetting,
#     SupportsFactoryRegistry,
#     FactoryBlueprint,
#     FactoryImplementable,
#     ToolFactoryImplementable,
#     is_blueprint,
#     is_valid_arg_setting,
#     default_is_arg_type,
#     convert_exp_def_to_string,
#     build_tree)

# from .results import (
#     METRICS_DATA_FOLDER,
#     METRICS_FOLDER,
#     METRICS_IMG_FOLDER,
#     collect_batch_results,
#     FolderManager)

# from .defaults import (
#     DEFAULT_JOURNAL_REPORTS_DIR,
#     DEFAULT_SETTINGS,
#     DEFAULT_BATCH_DIR,
#     DEFAULT_DATA_DIR,
#     DEFAULT_EXP_DEF_DIR,
#     DEFAULT_FILE_FORMAT)

from .exceptions import UnableToMeasureException



"""Measurement set is doing the heavy lifting of organization of
measurements. This module supports this set operations.
"""


import logging
import os
import csv
from urllib.parse import urljoin
import toml
from typing import (
    Union,
    List,
    Literal,
    Any,
    Sequence)
from adgtk.common import (
    ComponentDef,
    FactoryBlueprint,
    ArgumentSetting,
    FolderManager,
    InsufficientData,
    ArgumentType)
from adgtk.tracking import MetricTracker
from adgtk.factory.component import ObjectFactory
from adgtk.journals import ExperimentJournal
from adgtk.components.data import SupportsFiltering
from adgtk.utils.records import convert_record_as, process_as
from .base import (
    MeasInputType,
    Measurement,
    RecordAs,
    SupportsStopwords,
    Comparison,
    InvalidMeasurementConfiguration)


# ----------------------------------------------------------------------
#
# ----------------------------------------------------------------------
# py -m pytest -s test/instrumentation/test_meas_set.py


# ----------------------------------------------------------------------
# Module Constants
# ----------------------------------------------------------------------
DEBUG_TO_CONSOLE = False
LOW_DATA_THRESHOLD: Union[int, None] = None  # Set to None to disable


# ----------------------------------------------------------------------
# MeasurementSet
# ----------------------------------------------------------------------


class MeasurementSet:
    """Provides a protocol for a pre-defined grouping of measurements
    """
    description = "A measurement set which is a group of measurements."
    blueprint: FactoryBlueprint = {
        "group_label": "measurement-set",
        "type_label": "basic-set",
        "arguments": {
            "name": ArgumentSetting(
                argument_type=ArgumentType.STRING,
                default_value="main",
                help_str="The name of the measurement-set"),            
            "expected_result": ArgumentSetting(
                help_str="The result type. valid options are: number, list, string",
                default_value="number",
                argument_type=ArgumentType.STRING),
            "record_as": ArgumentSetting(
                help_str="The results should be saved as an avg, sum, latest, or raw",
                default_value="avg",
                argument_type=ArgumentType.STRING),
            "measurement_type": ArgumentSetting(
                help_str="The measurement type",
                default_value="single",
                argument_type=ArgumentType.STRING),
            "measurement_def": ArgumentSetting(
                argument_type=ArgumentType.LIST,
                help_str="A measurement that are part of this set",
                list_intro="This section will allow you to define the measurements for the measurement set.",
                list_min=1,
                list_arg_type=ArgumentType.BLUEPRINT,
                list_group_label="measurement"),
                        
            # TODO: Future release. Roadmap item.
            # "filters": ArgumentSetting(
            #     help_str="Filters for measuring 'placeholder only. hit enter.'",
            #     argument_type=ArgumentType.LIST,
            #     list_arg_type=ArgumentType.STRING),
        }
    }

    def __init__(
        self,
        factory: ObjectFactory,
        journal: ExperimentJournal,
        name: str,
        measurement_def: Union[List[ComponentDef], None],
        expected_result: Literal["number", "list", "string"],
        record_as: Literal["sum", "avg", "latest", "raw"],
        measurement_type: Literal["single", "compare", "dataset"],
        # filters: List[SupportsFiltering],
        create_metric_tracker: bool = True,        
        engine_name: str = "main"
    ) -> None:
        """MeasurementSet ensures that the engine can properly
        interact with the individual measurements.

        :param factory: The object factory
        :type factory: ObjectFactory
        :param journal: The Experiment journal
        :type journal: ExperimentJournal
        :param name: The unique name of the measurement set
        :type name: str
        :param measurement_type: The type of measurement
        :type measurement_type: Literal[single, compare,,dataset]
        :param director_override_blueprint: The director used for determing whether to run a measurement against a data type.
        :type director_override_blueprint: Union[dict, None]
        :param expected_result: The type of result expected from the measurement. Used to further filter measurements to run.
        :type expected_result: Literal[float, list, string, int]
        :param record_as: How should they be presented as output?
        :type record_as: Literal[sum;, avg, latest]
        :param measure_only: Provides filtering of data type
        :type measure_only: Union[List[Union[str, None]], None]
        :param filters: A filter of the data, defaults to None
        :type filters: Union[ List[Union[SupportsFiltering, Callable, None]], None], optional
        :param measurement_def: The measurements to perform,
            defaults to None
        :type measurement_def: Union[ List[dict], None], optional
        :param engine_name: The engine it supports, optional, defaults to main
        :type engine_name: str

        """
        self.name = name
        self.engine_name = engine_name
        self._journal = journal
        self.measurement_type = measurement_type  # lets the user update
        self.expected_result = expected_result  # lets the user update

        # convert and setup record_as
        try:
            self.record_as: RecordAs = convert_record_as(record_as)
        except InvalidMeasurementConfiguration as e:
            msg = f"meas-set: {name} has invalid record_as: {record_as}"
            logging.error(msg)
            raise InvalidMeasurementConfiguration(msg) from e

        # TODO: Roadmap
        # self.filters: List[SupportsFiltering] = []
        # if filters is not None:
        #     for candidate in filters:
        #         if isinstance(candidate, SupportsFiltering):
        #             self.filters.append(candidate)

        self.execution_count = 0

        # create measurements
        self._measurements: List[
            Comparison | Measurement] = []
        if measurement_def is not None:
            for meas_def in measurement_def:
                self._measurements.append(factory.create(meas_def))

        # and track. Set to create a metric tracker or expect to be
        # set post init for using a shared metric tracker?
        self.metric_tracker: Union[MetricTracker, None] = None
        self._metric_tracker_labels_created = False
        if create_metric_tracker:
            # create a metric tracker specific to this meas set
            self.metric_tracker = MetricTracker()
            self._init_metrics_into_metric_tracker()

    def _create_html_link(
        self,
        description: str,
        file_w_path: str,
        base_url: str = "http://127.0.0.1:8000"
    ) -> str:
        url_str = urljoin(
            base=base_url,
            url=file_w_path)
        return f"<a href={url_str}>{description}</a>"

    def _write_csv_and_metadata(
        self, metric: str,
        folders: FolderManager,
        base_url: str = "http://127.0.0.1:8000"
    ) -> str:
        """Performs the file processing of a single metric

        :param metric: The metric to export
        :type metric: str
        :param folders: The folders to save to local disk
        :type folders: FolderManager

        :return: HTML that can be used in a report.
        :rtype: str
        """
        html = ""
        file_prefix = f"{self.engine_name}.{self.name}"
        # metadata first
        meta_file = os.path.join(
            folders.metrics_data,
            f"{file_prefix}.meta.toml")

        # first metadata
        if self.metric_tracker is not None:
            meta = self.metric_tracker.get_metadata(metric)
            if len(meta) > 0:
                html += "<ul><li> Metadata "
                html += self._create_html_link(
                    description=meta_file,
                    file_w_path=meta_file,
                    base_url=base_url)

                html += "</li>"
                with open(meta_file, "w", encoding="utf-8") as meta_out:
                    if self.metric_tracker is not None:
                        toml.dump(meta, meta_out)

        # now data -> CSV
        if self.metric_tracker is not None:
            data = self.metric_tracker.get_all_data(metric)
            if len(data) > 0:
                html += "<li> Data file "
                data_file = os.path.join(
                    folders.metrics_data,
                    f"{file_prefix}.{metric}.csv")
                html += self._create_html_link(
                    description=data_file,
                    file_w_path=data_file,
                    base_url=base_url)
                html += "</li><ul>"

                with open(data_file, "w", encoding="utf-8") as data_out:
                    writer = csv.writer(data_out)
                    writer.writerow(data)

            elif self.metric_tracker is not None:
                logging.info(
                    f"Metric {metric} had zero data in set {self.name}")
        else:
            html += "<li> No Metric tracker found </li>"

        return html

    def create_html_and_export(
        self,
        folders: FolderManager,
        header: int = 2,
        base_url: str = "http://127.0.0.1:8000"
    ) -> str:
        """Creates local files and returns HTML that can be used by the
        reports. In addition, exports data, builds images, etc in order
        to save and report.

        :param folders: The folders to save to local disk
        :type folders: FolderManager
        :param header: The header, defaults to 2
        :type header: int, optional

        :return: HTML that can be used in a report.
        :rtype: str
        """

        # metadata & data
        if self.metric_tracker is None:
            return ""

        html = f"""
        <div>
            <h{header}>Measurement set {self.name}</h{header}>
        """

        meas_header = header+1

        for meas in self._measurements:
            # intro
            html += meas.report(header=meas_header)
            # image
            f_prefix = f"{self.engine_name}.{self.name}.{meas.tracker_label}"
            try:
                img_file = self.metric_tracker.line_plot(
                    label=meas.tracker_label,
                    folder=folders.metrics_img,
                    file_prefix=f_prefix)

                url_str = urljoin(
                    base=base_url,
                    url=img_file)

                html += f'<img src="{url_str}" alt="{meas.tracker_label}">'
            except InsufficientData:
                # NO-OP. unable to plot so not adding HTML for the IMG.
                pass

            # data
            html += self._write_csv_and_metadata(
                metric=meas.tracker_label,
                folders=folders,
                base_url=base_url)

        html += f"</div>"
        return html

    def register_metric_tracker(self, metric_tracker: MetricTracker) -> None:
        """Registers metric labels into a metric_tracker.

        :param metric_tracker: The metric tracker to register to
        :type metric_tracker: MetricTracker
        """
        self.metric_tracker = metric_tracker
        self._init_metrics_into_metric_tracker()

    def _init_metrics_into_metric_tracker(self) -> None:
        """Supports a first load of a set of metrics into the metric
        tracker. Used on init when create_metric_tracker is True or when
        running measurements for the first time the flag
        _metric_tracker_labels_created is False.

        Note: A MeasurementSet should not change, the design is that the
        Engine will add another set versus modify an existing one! So,
        This flag need only be set and checked once.
        """
        # safety first!
        if self.metric_tracker is None:
            raise InvalidMeasurementConfiguration("metric_tracker not set")

        # TODO: Consider moving from bool to raise Exception when an
        # entry exists. not needed for MVP.
        for meas in self._measurements:
            # registers using the component def set tracker_label
            ack = self.metric_tracker.register_metric(label=meas.tracker_label)
            if not ack:
                msg = f"{self.name} failed to register {meas.tracker_label}"
                logging.warning(msg)

        # and only perform once
        self._metric_tracker_labels_created = True

    def update_stopwords(self, stopwords: List[str]) -> None:
        """Updates every measurement that relies on stopwords.

        :param stopwords: The stopwords used by measurements.
        :type stopwords: List[str]
        """
        # NOTE: potential revisit this pattern. by having each measure
        # have a copy of the STOPWORDS it may be memory inefficient. The
        # benefit though is that each one is isolated and there should
        # be a minimum number of measurements.

        # update each measurement that supports stopwords
        for meas in self._measurements:
            if meas.features.can_use_stopwords:
                if isinstance(meas, SupportsStopwords):
                    if meas.use_stopwords:
                        meas.stopwords = stopwords

    def _process_measurement_w_data(
        self,
        data: Any,
        meas: Union[
            Measurement,
            Comparison],
        perform_compare_split: bool = False
    ) -> None:
        """Processes measurements

        :param data: The data to measure
        :type data: Any
        :param meas: The measurement
        :type meas: Union[ Measurement, Comparison]
        :param perform_compare_split: do a split?, defaults to False
        :type perform_compare_split: bool, optional
        :raises InvalidMeasurementConfiguration: missing tracker
        :raises InvalidMeasurementConfiguration: unknown split
        :raises InvalidMeasurementConfiguration: uknown type
        """

        # safety first
        if self.metric_tracker is None:
            raise InvalidMeasurementConfiguration("Missing metric_tracker")

        results = []
        if (isinstance(meas, Comparison)):
            if perform_compare_split:
                for row in data:
                    results.append(meas.compare(a=row[0], b=row[1]))
            else:
                # should be checked before invoking this method.
                msg = f"Unknown how to split for comparison {type(data)}"
                logging.error(msg)
                raise InvalidMeasurementConfiguration(msg)

        elif (isinstance(meas, Measurement)):
            for row in data:
                results.append(meas.measure(row))
        else:
            # should be checked before invoking this method.
            msg = f"Unknown measurement type: {type(meas)}"
            logging.error(msg)
            raise InvalidMeasurementConfiguration(msg)

        # safety
        if len(results) != len(data):
            # Don't store results. There is an issue
            msg = f"Failed to properly measure {meas.tracker_label}"
            logging.error(msg)
            return None

        if self.record_as == RecordAs.SUM:
            self.metric_tracker.add_data(
                label=meas.tracker_label,
                value=sum(results))
        elif self.record_as == RecordAs.AVG:
            self.metric_tracker.add_data(
                label=meas.tracker_label,
                value=sum(results) / len(results))
        elif self.record_as == RecordAs.AVG:
            self.metric_tracker.add_data(
                label=meas.tracker_label,
                value=results[-1])
        elif self.record_as == RecordAs.RAW:
            self.metric_tracker.add_raw_data(
                label=meas.tracker_label,
                values=results)
        else:
            msg = f"Unknown value of record_as: {self.record_as}"
            logging.error(msg)
            raise InvalidMeasurementConfiguration(msg)

    def perform_measurements(self, data: Sequence) -> None:
        """Executes the measurements if valid. Increments the
        execution_count if measurements are performed.

        :param data: the data to measure
        :type data: Any
        """
        # TODO: code filters
        # if len(self.filters) > 0:
        #     raise NotImplementedError("DEVELOPMENT NEEDED")

        if self.metric_tracker is None:
            raise InvalidMeasurementConfiguration("metric_tracker not set")

        if LOW_DATA_THRESHOLD is not None:
            if len(data) <= LOW_DATA_THRESHOLD:
                msg = f"{self.name}: low data threshold alert on measure"
                logging.warning(msg)

        # ensure metric tracker is setup (needed with delayed setup)
        if not self._metric_tracker_labels_created:
            self._init_metrics_into_metric_tracker()

        # Get data type
        input_type = process_as(data)
        perform_compare_split = False
        if input_type == MeasInputType.PAIR_AS_LIST:
            perform_compare_split = True

        update_counter = False
        for meas in self._measurements:
            # execute measurement
            if input_type[0] in meas.features.input_type or \
                    input_type[1] in meas.features.input_type:
                update_counter = True
                self._process_measurement_w_data(
                    data=data,
                    meas=meas,
                    perform_compare_split=perform_compare_split)
            elif DEBUG_TO_CONSOLE:
                print(f"Skipping {meas.tracker_label} : "
                      f"{input_type[0]} | {input_type[1]}")

        # wrap-up. Increment the counter.
        if update_counter:
            self.execution_count += 1



"""For ease of working with the results folder. This module provides
a class to manage the folder structure for an experiment.
"""

import logging
import os
import shutil
from typing import Union
from adgtk.utils import load_settings

# ----------------------------------------------------------------------
#
# ----------------------------------------------------------------------
# py -m pytest -s test/common/test_results.py

# ----------------------------------------------------------------------
# CONSTANTS
# ----------------------------------------------------------------------
METRICS_FOLDER = "metrics"
METRICS_IMG_FOLDER = "images"
METRICS_DATA_FOLDER = "data"
DATASET_FOLDER = "datasets"
AGENT_FOLDER = "agent"
PERFORMANCE_FOLDER = "performance"

# ----------------------------------------------------------------------
# Functions
# ----------------------------------------------------------------------


class FolderManager:
    """A simple class to help manage and ensure a consistent folder
    structure for an experiment results. It takes the experiment name
    and the settings file to create the necessary folders and
    sub-folders. It also provides an easy way to access the paths to
    the respective use by reading the attributes.
    
    The class has the following useful attributes:
        * base_folder: the root folder for the experiment
        * agent: the folder for agent data
        * metrics: the folder for metrics
        * metrics_data: the folder for metrics data
        * metrics_img: the folder for metrics images
        * dataset: the folder for datasets
        * performance: the folder for performance data
        * model_dir: the folder for models
    """

    def __init__(
        self,
        name: str,
        settings_file_override: Union[str, None] = None,
        clear_and_rebuild: bool = False
    ):
        """Initialize and create if needed

        :param name: the experiment name
        :type name: str
        :param clear_and_rebuild: deletes all results, defaults to False
        :type clear_and_rebuild: bool
        :param settings_file_override: redirect settings file location,
            defaults to None
        :type settings_file_override: Union[str, None], optional
        :raises FileNotFoundError: Unable to load the settings file
        """

        # cleanup to ensure consistency
        name = name.lower()
        if name.endswith(".toml") or name.endswith(".yaml"):
            name = name[:-5]

        # now load the settings and set the items to manage
        try:
            settings = load_settings(file_override=settings_file_override)
        except FileNotFoundError as e:
            if settings_file_override is not None:
                msg = f"Unable to locate settings file {settings_file_override}"
            else:
                msg = "Unable to locate settings file from default location"
            logging.error(msg)
            raise FileNotFoundError(msg) from e

        exp_result_dir = os.path.join(settings.experiment["results_dir"], name)
        self.base_folder = exp_result_dir

        if os.path.exists(exp_result_dir) and clear_and_rebuild:
            shutil.rmtree(exp_result_dir)

        if not os.path.exists(exp_result_dir):
            os.makedirs(exp_result_dir, exist_ok=True)
            logging.info(f"Created {exp_result_dir}")
        else:
            msg = f"Folder Manager found {exp_result_dir}. No action taken "\
                  "to create additional folders"
            logging.info(msg)

        # Agent data
        self.agent = os.path.join(exp_result_dir, AGENT_FOLDER)
        os.makedirs(self.agent, exist_ok=True)

        # Create metrics sub-folder(s)
        self.metrics = os.path.join(exp_result_dir, METRICS_FOLDER)
        self.metrics_data = os.path.join(
            self.metrics, METRICS_DATA_FOLDER)
        self.metrics_img = os.path.join(self.metrics, METRICS_IMG_FOLDER)

        os.makedirs(self.metrics, exist_ok=True)
        os.makedirs(self.metrics_data, exist_ok=True)
        os.makedirs(self.metrics_img, exist_ok=True)

        # create dataset folder
        self.dataset = os.path.join(exp_result_dir, DATASET_FOLDER)
        os.makedirs(self.dataset, exist_ok=True)

        # Performance
        self.performance = os.path.join(exp_result_dir, PERFORMANCE_FOLDER)
        os.makedirs(self.performance, exist_ok=True)

        # Models - added 0.1.1a1
        self.model_dir = "models"
        try:
            self.model_dir = settings.model_dir
        except AttributeError:
            logging.warning("Using older settings file. No model_dir found")
            self.model_dir = os.path.join(exp_result_dir, "models")

        if not os.path.exists(self.model_dir):
            os.makedirs(self.model_dir, exist_ok=True)

        # and additional/existing files/folders
        self.logfile = os.path.join(settings.logging["log_dir"], f"{name}.log")
        self.data_dir = settings.experiment["data_dir"]
        self.tensorboard_dir = settings.experiment["tensorboard_dir"]

    def __str__(self) -> str:
        """Creates a string representation of the FolderManager object

        :return: a useful string for UX via a CLI.
        :rtype: str
        """
        to_string = "FolderManager\n"
        to_string += "-------------\n"
        to_string += f" - log: {self.logfile}\n"
        to_string += f" - tensorboard: {self.tensorboard_dir}\n"
        to_string += f" - models: {self.model_dir}\n"
        to_string += f" - metrics: {self.metrics}\n"
        to_string += f"    - data: {self.metrics_data}\n"
        to_string += f"    - images: {self.metrics_img}\n"
        to_string += f" - dataset: {self.dataset}\n"

        return to_string



"""For ease of working with the results folder. This module provides
a class to manage the folder structure for an experiment.

Logging
=======
1. uses the same logger file as experiment.runner. adgtk.runner.log

Testing
=======
py -m pytest -s test/common/test_results.py

"""

# setup logfile for this and sub-modules
from adgtk.tracking.structure import (
    EXP_DATASET_FOLDER,
    EXP_METRICS_FOLDER,
    EXP_IMG_FOLDER,
    EXP_MODEL_DIR,
    EXP_OTHER_DIR,
    EXP_RESULTS_FOLDER,
    EXP_MODEL_TRAIN_LOG,
    ExperimentRunFolders
)
from adgtk.common.defaults import (
    BATCH_DEF_DIR,
    EXP_DEF_DIR,
    LOG_DIR,
    SHARED_MODEL_DIR)
import os
from adgtk.utils import create_logger

# Set up module-specific logger
_logger = create_logger(
    "adgtk.runner.log",
    logger_name=__name__,
    subdir="framework"
)


# ----------------------------------------------------------------------
# CONSTANTS
# ----------------------------------------------------------------------
DEBUG = False

CONCLUSION = "conclusions"

# ----------------------------------------------------------------------
# Helper functions
# ----------------------------------------------------------------------


def _build_run_dirs(root_dir: str, experiment_name) -> ExperimentRunFolders:
    _logger.info("Building run folders at %s", root_dir)
    
    return ExperimentRunFolders(
        datasets=os.path.join(root_dir, EXP_DATASET_FOLDER),
        metrics=os.path.join(root_dir, EXP_METRICS_FOLDER),
        images=os.path.join(root_dir, EXP_IMG_FOLDER),
        other=os.path.join(root_dir, EXP_OTHER_DIR),
        conclusion=os.path.join(root_dir, CONCLUSION),
        root_dir=root_dir,
        log_dir=os.path.join("logs", "runs", experiment_name),
        experiment_name=experiment_name,
        common=os.path.join(EXP_RESULTS_FOLDER, experiment_name, "common"),
        model_dir=os.path.join(root_dir, EXP_MODEL_DIR),
        train_log_dir=os.path.join(root_dir, EXP_MODEL_TRAIN_LOG)
    )


def _get_run_dir(experiment_name: str, run_id: str, build: bool = True) -> str:

    # Experiment
    exp_folder = os.path.join(EXP_RESULTS_FOLDER, experiment_name)
    # RUN
    run_dir = os.path.join(exp_folder, run_id)
    common_dir = os.path.join(exp_folder, "common")
    if build:
        os.makedirs(EXP_RESULTS_FOLDER, exist_ok=True)
        os.makedirs(exp_folder, exist_ok=True)
        os.makedirs(run_dir, exist_ok=True)
        os.makedirs(common_dir, exist_ok=True)
    return run_dir
# ----------------------------------------------------------------------
# Functions
# ----------------------------------------------------------------------

# TODO: Broken


def collect_batch_results(exp_prefix: str, results_dir: str) -> list:
    """Retrieves the results and the configuration of an experiment and
    puts it into a dictionary for processing.

    :param experiment_name: The name of the experiment
    :type experiment_name: str
    :return: a tuple of filename and a dict that contains results
    :rtype: dict
    """
    raise NotImplementedError("collect_batch_results() requires an update")
    experiments = os.listdir(results_dir)
    results = []
    # for experiment in experiments:
    #     if experiment.startswith(exp_prefix):
    #         folder_manager = ExperimentFolderManager(experiment)
    #         results.append((experiment, folder_manager.collect_results()))

    return results


def build_project_folders(base_path: str = ""):
    """Builds all the required folders for a project."""
    def make_dir(path):
        full_path = os.path.join(base_path, path) if base_path else path
        os.makedirs(full_path, exist_ok=True)
        _logger.info(f"Created folder: {full_path}")
    make_dir(BATCH_DEF_DIR)
    make_dir(EXP_DEF_DIR)
    make_dir(SHARED_MODEL_DIR)
    make_dir(LOG_DIR)
    make_dir(EXP_RESULTS_FOLDER)


def build_folder_listing(
    experiment_name: str,
    run_id: str
) -> ExperimentRunFolders:
    """Verifies the folders exist and returns an easy to use object for
    use within different code bases.

    :param experiment_name: The experiment name
    :type experiment_name: str
    :param run_id: The id of the individual run
    :type run_id: str
    :return: a Pydantic Model that provides ease of use
    :rtype: ExperimentRunFolders
    """
    root_dir = _get_run_dir(
        experiment_name=experiment_name, run_id=run_id, build=False)
    
    exp_dir_listing = _build_run_dirs(
        root_dir=root_dir,
        experiment_name=experiment_name)
    folders = exp_dir_listing.to_dict()
    missing = False
    for key, dir in folders.items():
        if not os.path.exists(dir) and not key == "experiment_name" \
            and key == "common":
            msg = f"missing folder: {dir}"
            _logger.info(msg)
            missing = True

    if missing:
        raise FileNotFoundError("Missing one or more folders. Check log.")
    return exp_dir_listing


def setup_run(experiment_name: str, run_id: str) -> ExperimentRunFolders:
    """Creates all the required folders for results of an experiment run

    :param experiment_name: The name of the experiment
    :type experiment_name: str
    :param run_id: The unique idenfier of a run of an experiment
    :type run_id: str
    :return: An easy to use object for referring to different folders
        for the run such as the datasets folder.
    :rtype: ExperimentRunFolders
    """
    # ensure results folder exists
    root_dir = _get_run_dir(
        experiment_name=experiment_name, run_id=run_id, build=True)

    exp_dir_listing = _build_run_dirs(
        root_dir=root_dir,
        experiment_name=experiment_name)
    folders = exp_dir_listing.to_dict()
    # Builds the subfolders, etc.
    for key, folder in folders.items():
        if key != "experiment_name" :
            os.makedirs(folder, exist_ok=True)

    # now verify and create ExperimentRunFolders
    listing = build_folder_listing(
        experiment_name=experiment_name,
        run_id=run_id
    )
    _logger.info(f"Setup folders using schema version {listing.version}")
    return listing



"""constants.py handle Constants for the CLI which are larger than
simple strings"""


BOOT_FILENAME = "bootstrap.py"

BOOT_PY = """
\"""bootstrap.py

This script is your startup script. Use this to load your custom code
into the factory.
\"""

import adgtk.factory.component as factory
from adgtk.builtins.scenario import HelloWorldScenario

def foundation():
    \"""These are built-in components that should not be removed unless
    you are replacing them with another. loss of these will impact
    overall operations.
    \"""
    pass
    # factory.register()

def builtin():
    \"""these are optional components that can be removed if desired.\"""
    factory.register(HelloWorldScenario)
    
def user_code():
    pass
    # factory.register()
"""



"""runner.py is focused on running and managing a scenario. A Scenario
is the root of any experiment.

TODO
====

1. finalize design for how to handle results
2. log results, experiment run, etc
3. Test nested definitions

Defects
=======
1. log to console is not printing. observed w/nohop at least, need to TS
"""

import datetime
import os
import sys
# before importing others
# ----------------------------------------------------------------------
# Start of path verification
# ----------------------------------------------------------------------
path = os.getcwd()
bootstrap_file = os.path.join(path, "bootstrap.py")
if not os.path.exists(bootstrap_file):
    print("ERROR: Unable to locate the bootstrap.py. Please check your path.")
    sys.exit(1)
# ----------------------------------------------------------------------
# End of path verification
# ----------------------------------------------------------------------

# setup logfile for this and sub-modules
from adgtk.utils import create_logger
from logging import Logger
import inspect
from typing import Optional, Union
import yaml
from pydantic import ValidationError
from adgtk.common.defaults import (
    EXP_DEF_DIR,
    BATCH_DEF_DIR,
    SCENARIO_LOGGER_NAME
)
from adgtk.experiment.structure import (
    AttributeEntry,
    ExperimentDefinition,
    ScenarioResults,
    ScenarioProtocol,
    BuildComponentResult,
)
import adgtk.factory.component as factory
import adgtk.tracking.project as project_manager
import adgtk.tracking.journal as exp_journal
from adgtk.tracking.structure import (
    AvailableExperimentModel,
    ExperimentRunFolders)
from adgtk.tracking.utils import setup_run
from adgtk.utils import get_user_input
from adgtk.experiment.structure import BatchDefinition


# ----------------------------------------------------------------------
# Constants
# ----------------------------------------------------------------------
RESULTS_FILE = "results.yaml"
RUN_CONFIG_FILE = "run.exp.config.yaml"
TO_CONSOLE = False                           # for development only

# ----------------------------------------------------------------------
# Logging. target log varies by need
# ----------------------------------------------------------------------
_logger: Optional[Logger] = None

# ----------------------------------------------------------------------
# Helper functions
# ----------------------------------------------------------------------


def _contains_blueprint_dir(path: str) -> bool:
    """helps confirm how to format the filename by checking a string for
    whether the path contains the experiment directory "blueprints"

    :param path: The path to inspect
    :type path: str
    :return: True if the path already points to the blueprints folder
    :rtype: bool
    """

    # Normalize path to avoid issues with slashes
    parts = os.path.normpath(path).split(os.sep)
    return EXP_DEF_DIR in parts


def _contains_batch_dir(path: str) -> bool:
    """helps confirm how to format the filename by checking a string for
    whether the path contains the experiment directory "batch"

    :param path: The path to inspect
    :type path: str
    :return: True if the path already points to the batch folder
    :rtype: bool
    """
    parts = os.path.normpath(path).split(os.sep)
    return BATCH_DEF_DIR in parts


def _build_component(
    attribute_def: AttributeEntry,
) -> BuildComponentResult:
    """Builds a component. This is designed to be recursive.

    :param attribute_def: The configuration to build
    :type attribute_def: AttributeEntry
    :raises ValueError: Occurs when unexpected combinations occur
    :return: At the root its Callable, else its an argument.
    :rtype: Union[Callable, float, bool, str, int, list, None]
    """

    # setup
    if isinstance(attribute_def, dict):
        try:
            attribute_def = AttributeEntry(**attribute_def)
        except:
            msg = f"Invalid enty type: {attribute_def}"
            if _logger is not None:
                _logger.error(msg)
            raise ValueError(msg)
    if isinstance(attribute_def.init_config, (bool, int, str, float)):
        msg = (f"Attribute: {attribute_def.attribute} = "
               f"{attribute_def.init_config}")
        if _logger is not None:
            _logger.info(msg)
        return attribute_def.init_config

    # 1. ensure proper type alignment
    if not isinstance(attribute_def, AttributeEntry):
        try:
            attribute_def = AttributeEntry(**attribute_def)
        except ValidationError:
            msg = f"Corrupt definition: {attribute_def}"
            raise ValueError(msg)

    # 1.5 if factory does not init
    factory_id = attribute_def.factory_id
    if factory_id is None:
        msg = f"Corrupt definition: {attribute_def}"
        if _logger is not None:
            _logger.info(msg)
        raise ValueError(msg)

    if not attribute_def.factory_init:
        msg = f"Getting creator not init for {factory_id}"
        if _logger is not None:
            _logger.info(msg)

        return factory.get_callable(factory_id)

    # 2. assemble init values (if going to init)
    args = {}
    init_config = attribute_def.init_config
    # in the event factory can init but there is nothing to config and
    # the user code does not set init_config.
    if init_config is None:
        init_config = []

    if init_config is None:
        msg = f"Corrupt definition: {attribute_def}. Unexpected None"
        if _logger is not None:
            _logger.info(msg)

        raise ValueError(msg)

    if not isinstance(init_config, list):
        msg = f"Corrupt definition: {attribute_def}. Expected init_config list"
        if _logger is not None:
            _logger.info(msg)
        raise ValueError(msg)

    for item in init_config:
        if isinstance(item, dict):
            try:
                item = AttributeEntry(**item)
            except ValidationError:
                msg = f"Unexpected dict: {init_config}"
                if _logger is not None:
                    _logger.info(msg)
                raise ValueError(msg)

        # and final check
        if isinstance(item, AttributeEntry):
            args[item.attribute] = _build_component(item)

    result = factory.create(factory_id=factory_id, **args)
    return result   # type: ignore


def _load_scenario(
        exp_config: ExperimentDefinition) -> ScenarioProtocol:
    """A specialized helper function that performs safety checks before
    using the _build_component to build out the scenario.

    :param exp_config: The configuration to load
    :type exp_config: ExperimentDefinition
    :raises ValueError: Unexpected values for the configuration
    :return: A created scenario
    :rtype: Callable
    """

    # Safety checks
    if not isinstance(exp_config, ExperimentDefinition):
        try:
            exp_config = ExperimentDefinition(**exp_config)
        except ValidationError:
            msg = "Invalid experiment definition"            
            raise ValueError(msg)

    scenario_config = exp_config.init_config

    if not isinstance(scenario_config, AttributeEntry):
        msg = f"Invalid Scenario definition. got {type(scenario_config)}"
        if _logger is not None:
            _logger.info(msg)
        raise ValueError(msg)
    # now build
    msg = f"Attempting to load Scenario: {scenario_config}"
    if _logger is not None:
        _logger.info(msg)
    scenario = _build_component(scenario_config)

    if not isinstance(scenario, ScenarioProtocol):
        msg = (
            "Object defined does not implement the Scenario Protocol. "
            f"Type found is {type(scenario)}"
        )
        if _logger is not None:
            _logger.error(msg)
        raise ValueError(msg)

    # and one last safety check
    if inspect.isclass(scenario):
        msg = f"failed to create scenario. created: {type(scenario)}"
        if _logger is not None:
            _logger.info(msg)
        raise ValueError(msg)

    return scenario


def _load_experiment_file(filename: str) -> ExperimentDefinition:
    """loads an experiment file

    :param filename: The file to load
    :type filename: str
    :return: the configuration needed to build the experiment.
    :rtype: ExperimentDefinition
    """
    # get config
    if not _contains_blueprint_dir(filename):
        file_w_path = os.path.join("blueprints", filename)
    else:
        file_w_path = filename

    if not file_w_path.endswith(".yaml"):
        file_w_path += ".yaml"

    with open(file=file_w_path, mode="r", encoding="utf-8") as infile:
        try:
            exp_def = yaml.safe_load(infile)
        except yaml.YAMLError:
            msg = f"Malformed file: {file_w_path}"
            if _logger is not None:
                _logger.info(msg)
            raise ValueError(msg)

    # and convert for safety and validation
    try:
        exp_config = ExperimentDefinition(**exp_def)
    except ValidationError:
        msg = f"Corrupted experiment configuration: {filename}"
        if _logger is not None:
            _logger.info(msg)
        sys.exit(1)

    return exp_config


def _select_experiment() -> str:
    """Interacts with a user to determine which experiment to run

    :raises ValueError: Unexpected type. Internal processing failure
    :raises RuntimeError: Internal processing failure
    :return: the filename for the experiment to run
    :rtype: str
    """
    choices = []
    idx_to_choice_map = []

    available_list = project_manager.get_available_experiments()
    if len(available_list) == 0:
        msg = "No available experiments found."
        if _logger is not None:
            _logger.info(msg)
        print(msg)
        sys.exit(0)

    name_str = "name"
    desc_str = "description"

    title = f"{name_str:<25} | {desc_str}"
    bar_length = len(title)
    entries_str = ""
    for idx, entry in enumerate(available_list):
        if not isinstance(entry, AvailableExperimentModel) and \
                isinstance(entry, dict):
            try:
                entry = AvailableExperimentModel(**entry)
            except ValidationError:
                msg = ("Unexpected type from project_manager for exp "
                       f"listing. Observed: {type(entry)}")
                if _logger is not None:
                    _logger.info(msg)
                raise ValueError(msg)

        entry_str = f" {idx} : {entry.name:<22} | {entry.description}"
        if len(entry_str) > bar_length:
            bar_length = len(entry_str)
        entries_str += f"{entry_str}\n"
        choices.append(entry.name)
        choices.append(str(idx))
        idx_to_choice_map.append(entry.name)

    bar = "="*bar_length
    output_str = f"\n{title}\n{bar}\n{entries_str}\n"
    print(output_str)
    exp_choice = get_user_input(
        user_prompt="Which experiment do you want to run",
        requested="str",
        choices=choices,
        helper="Based on current blueprints folder listing"
    )
    try:
        idx = int(exp_choice)
        return idx_to_choice_map[idx]
    except ValueError:
        if isinstance(exp_choice, str):
            return exp_choice

    if _logger is not None:
        _logger.info("Failed to determine experiment")
    raise RuntimeError("Get experiment failure")


def _save_copy_of_config(
    config: Union[ExperimentDefinition, dict],
    root_dir: str
) -> None:
    """Saves a copy of the configuration to the root of the run dir

    :param config: The configuration of the experiment being run
    :type config: Union[ExperimentDefinition, dict]
    :param root_dir: The root directory of the run
    :type root_dir: str
    :raises FileNotFoundError: The root_dir is not found
    """
    if not os.path.exists(root_dir):
        raise FileNotFoundError(root_dir)

    if isinstance(config, ExperimentDefinition):
        config = config.model_dump()

    file_w_path = os.path.join(root_dir, RUN_CONFIG_FILE)
    with open(file=file_w_path, mode="w", encoding="utf-8") as outfile:
        yaml.dump(config, outfile)


# ----------------------------------------------------------------------
# Public
# ----------------------------------------------------------------------
# Design detail: Only the public can set the _logger. This helps ensure
# a consistent behavior with the logger.

# TODO: update docstring
def run_scenario(
    filename: Optional[str] = None,
    append_timestamp:bool=False,
    use_count:bool=True,
    print_to_console:bool = True
) -> tuple[ScenarioResults, ExperimentRunFolders]:
    """runs a scenario based on the definition within a file

    :param filename: The file to load
    :type filename: str
    :return: The run results and where to find them
    :rtype: tuple[ScenarioResults, ExperimentRunFolders]
    """
    global _logger    
    exp_journal.reset() # cover any batch processing
    if filename is None:
        exp_name = _select_experiment()
        exp_name += ".yaml"
        filename = os.path.join(EXP_DEF_DIR, exp_name)
    config = _load_experiment_file(filename)
    run_id = project_manager.get_next_experiment_run_id(
        experiment_name=config.name,
        use_count=use_count,
        append_timestamp=append_timestamp,
        prefix=None
    )
    _logger = create_logger(
        logfile="scenario.log",
        logger_name=SCENARIO_LOGGER_NAME,
        subdir="runs",
        experiment_name=config.name,
        log_to_console=print_to_console or TO_CONSOLE
    )
    # setup the folders for the results
    folders = setup_run(experiment_name=config.name, run_id=run_id)
    # and save a copy for future reference
    _save_copy_of_config(config=config, root_dir=folders.root_dir)
    # so I  can log
    log_file = os.path.join("logs", "runs", config.name, "scenario.log")
    scenario = _load_scenario(config)
    _logger.info("-"*60)
    _logger.info("Starting Scenario")
    _logger.info("Results folder: %s", folders.root_dir)
    _logger.info("Starting logging at %s", log_file)
    _logger.info("-"*60)
    intro = f"| Starting experiment {config.name}: run {run_id} |"
    print("-"*len(intro))
    print(intro)
    print("-"*len(intro))
    result = scenario.run_scenario(result_folders=folders)
    exp_journal.save_journal(folders.conclusion)
    _logger.info("-"*60)
    _logger.info("Scenario Execution complete")
    _logger.info("-"*60)

    # ensure valid
    if not isinstance(result, ScenarioResults):
        try:
            result = ScenarioResults(**result)
        except ValidationError:
            msg = "Unable to convert Scenario result. Unable to update project"
            _logger.error(msg)
            print("ERROR: " + msg)

    # no central tracking of results. saving on-disk. less to centrally manage
    # and if in the future a listing is needed its safer to walk the disk.
    results_file_w_path = os.path.join(folders.conclusion, RESULTS_FILE)

    with open(file=results_file_w_path, mode="w", encoding="utf-8") as outfile:
        yaml.safe_dump(result.model_dump(), outfile)    
    return (result, folders)


def run_batch(filename: str, print_to_console:bool = True) -> None:
    """Runs a batch of experiments. 

    Future: Consider building a selector for batches

    :param filename: The filename to load
    :type filename: str
    :raises FileNotFoundError: unable to locate batch file
    """
    global _logger
    if _logger is None:
        _logger = create_logger(
            logfile="adgtk.runner.log",
            logger_name=__name__,
            subdir="framework",
            log_to_console=print_to_console or TO_CONSOLE
        )

    if not _contains_batch_dir(filename):
        filename = os.path.join(BATCH_DEF_DIR, filename)

    if not filename.endswith(".yaml"):
        filename += ".yaml"

    if not os.path.exists(filename):
        msg = f"Unable to find batch: {filename}"

        _logger.info(msg)

        raise FileNotFoundError(msg)

    with open(file=filename, mode="r", encoding="utf-8") as infile:
        data = yaml.safe_load(infile)

    try:
        batch = BatchDefinition(**data)

    except ValidationError:
        msg = f"Corrupt batch file: {filename}"
        if _logger:
            _logger.error(msg)
        print(msg)  # always print when this occurs
        sys.exit(1)

    # now run the experiments. serially of course.
    _logger.info("Starting batch run %s", batch.name)
    print(f"Starting batch run {batch.name}")
    for idx, experiment in enumerate(batch.experiments):
        print(f"{datetime.datetime.now()} Running Experiment: {experiment}: "
              f"{idx} of {len(batch.experiments)}")
        _logger.info("Batch running experiment %s. %d of %d",
                     experiment,
                     idx,
                     len(batch.experiments))
        
        _ = run_scenario(filename=experiment)



"""Model operations.
"""

from typing import Protocol, runtime_checkable, Iterable, Any

# ----------------------------------------------------------------------
# Protocols
# ----------------------------------------------------------------------


@runtime_checkable
class CanMoveToCPU(Protocol):
    """Ensures the model can move to the CPU"""

    def to_cpu(self):
        """Moves the model to the CPU"""


@runtime_checkable
class CanMoveToGPU(Protocol):
    """Ensures the model can move to the GPU"""

    def to_gpu(self):
        """Moves the model to the GPU"""


@runtime_checkable
class CanSaveModel(Protocol):
    """Ensures the model can be saved"""

    def save_model(self, save_folder: str) -> None:
        """Saves the model to disk

        :param save_folder: The folder where the model is saved
        :type save_folder: str
        """


@runtime_checkable
class CanLoadModel(Protocol):
    """Ensures the model can be loaded from disk"""

    def load_model(self, from_folder: str, **kwargs) -> None:
        """Loads the model from disk

        :param from_folder: The folder where the model is saved
        :type from_folder: str
        """


@runtime_checkable
class CanGenerateTextUsingPrompt(Protocol):
    """Provides a consistent interface for wrapping a LLM. This is not
    required when using a LLM, but reduces the amount of code when using
    built-in agents."""

    def generate(self, prompt: str, **kwargs) -> str:
        """Generates text based on the prompt

        :param prompt: the prompt for the model
        :type prompt: str
        :return: the resulting text
        :rtype: str
        """


@runtime_checkable
class TrainableModel(Protocol):
    """A trainable model"""

    def train(self, train_dataset: Iterable, **kwargs) -> None:
        """Trains the model

        :param train_dataset: The dataset to train with
        :type train_dataset: Iterable
        """

    def infer(self, sample: Any, **kwargs) -> Any:
        """Infers from a model

        :param sample: The sample to infer
        :type sample: Any
        :return: the model output
        :rtype: Any
        """



"""Similar to the Component factory, the Tool factory is responsible for
creating tools that can be used by Agents. The factory is designed to
make it easier to create new tools and to manage the tools that are
available to the Agent.
"""
from __future__ import annotations
import json
from typing import Any, Union, List
from adgtk.journals import ExperimentJournal
from adgtk.common import DuplicateFactoryRegistration, ToolFactoryImplementable

# py -m pytest -s test/factory/test_tool_factory.py

class ToolFactory:
    """The ToolFactory class is responsible for creating tools that can
    be used by Agents. The factory is designed to make it easier to
    create new tools and to manage the tools that are available to
    the Agent.
    """

    def __init__(
        self,
        journal: Union[ExperimentJournal, None] = None,
        factory_name: str = "Tool Factory"
    ) -> None:
        self.factory_name = factory_name
        self.journal = journal
        self._tools:dict[str, ToolFactoryImplementable] = {}

    def __len__(self) -> int:
        return len(self._tools)
    
    def __str__(self) -> str:
        """Basic report of the ToolFactory.

        :return: a report string of the ToolFactory.
        :rtype: str
        """
        title = "Tool Factory report"
        report = ""
        report += f"{title}\n"
        report += "---------------------\n"
        for key in sorted(self._tools.keys()):
            report += f"- {key}\n"    
        report += f"---------------------\n"
        return report

    def registry_listing(self) -> List[str]:
        """Lists the registered tools in the ToolFactory.

        :return: a list of usable tools.
        :rtype: List[str]
        """
        return list(self._tools.keys())
    
    def register(self, tool: ToolFactoryImplementable) -> None:
        tool_name = tool.definition["function"]["name"]
        if tool_name in self._tools:
            msg = f"Tool {tool_name} has already been registered "\
                  "with the ToolFactory."
            raise DuplicateFactoryRegistration(msg)
        
        if self.journal is not None:        
            self.journal.add_entry(
                entry_type="tool",
                entry_text=tool_name,
                component="tool_factory",
                include_timestamp=True)
        self._tools[tool_name] = tool

    def unregister(self, tool_name: str) -> None:
        if tool_name not in self._tools:
            msg = f"Tool {tool_name} has not been registered with "\
                  "the ToolFactory."
            raise KeyError(msg)

        del self._tools[tool_name]
        
    def get_tools_json(self) -> List[str]:
        """Returns a list of all the tools that have been registered
        with the ToolFactory. The primary purpose is to provide an easy
        method for creating a JSON representation of the tools for the
        language model to use (such as ollama via "tools").

        :return: The JSON representation of the tools.
        :rtype: List[str]
        """
        tool_list = []
        for key, tool in self._tools.items():
            tool_list.append(json.dumps(tool.definition))

        return tool_list
    
    def create(self, tool_name: str) -> Any:
        """Creates a new instance of the specified tool.

        :param tool_name: The name of the tool to create.
        :type tool_name: str
        :return: The new instance of the tool.
        :rtype: Any
        """
        if tool_name not in self._tools:
            msg = f"Tool {tool_name} has not been registered with the "\
                  "ToolFactory."
            raise KeyError(msg)

        tool = self._tools[tool_name]
        if isinstance(tool, ToolFactoryImplementable):            
            new_tool = tool()
            return new_tool
        
        raise TypeError("Tool is not a ToolFactoryImplementable object.")


"""builtin.py provides built-in scenario's for testing and user usage."""

from typing import Callable, ClassVar
from adgtk.experiment.structure import ScenarioResults
from adgtk.factory import SupportsFactory, BlueprintQuestion
from adgtk.tracking.structure import ExperimentRunFolders

# ----------------------------------------------------------------------
# Interview
# ----------------------------------------------------------------------
_q1 = BlueprintQuestion(
    attribute="delay",
    question="This is the delay before saying hello",
    entry_type="float",
    default_value=1
)

_q2 = BlueprintQuestion(
    attribute="measurement",
    question="Which measurement",
    entry_type="expand",
    group="measure"
)

hello_world_interview = [
    _q1
]
nested_interview = [
    _q1,
    _q2
]

# ----------------------------------------------------------------------
# Class
# ----------------------------------------------------------------------

class HelloWorldScenario(SupportsFactory):
    """This scenario prints hello world. It is useful for testing and
    development. More importantly it is useful for validating your
    environment when you get setup as well as being used as an example.
    """
    factory_id: ClassVar = "hello.world"
    group: ClassVar = "scenario"
    tags: ClassVar = ["demo", "development"]
    summary: ClassVar = "A simple scenario to validate your environment"
    interview_blueprint: ClassVar =  hello_world_interview
    factory_can_init: ClassVar = True

    def __init__(self, delay:float) -> None:        
        super().__init__()
        self.delay = delay


    def run_scenario(
        self,
        result_folders:ExperimentRunFolders
    ) -> ScenarioResults:
        print(f"Hello World from the Scenario! - attr={self.delay}")
        return ScenarioResults(files=[])

class NestedWorld(SupportsFactory):
    """A demo Scenario that will require multiple expansions
    """
    factory_id: ClassVar = "nested.world"
    group: ClassVar = "scenario"
    tags: ClassVar = ["demo", "development"]
    summary: ClassVar = "A nested scenario to validate your environment"
    interview_blueprint: ClassVar =  nested_interview
    factory_can_init: ClassVar = True

    def __init__(self, delay:float, measurement:Callable) -> None:        
        super().__init__()
        self.delay = delay        
        self.measure = measurement

    def run_scenario(
        self,
        result_folders:ExperimentRunFolders
    ) -> ScenarioResults:        
        print(f"Hello NestedWorld from the Scenario! - attr={self.delay}")
        print(f"MEAS: {self.measure}")
        return ScenarioResults(files=[])





"""structure.py

Goals
=====
Provides the structure that is factory specific.

Design
======
The focus on this structure is to reduce the amount of code needed to
work with the factory. For example the FactoryOrder provides an easy to
use structure for creating objects while the SupportsFactory reduces the
amount of code needed in custom classes, registration, etc.

Structures
==========
 - User-facing models: validated, serialized via BaseModel
 - Internal helpers: lightweight dataclasses, not validated


Testing
=====
N/A

Notes
=====
1. This takes the lessons learned from earlier adgtk versions to reduce
   the amount of code the user needs to write as well as simplifying the
   overall code maintenance for this project.
"""
from abc import ABC
import secrets
from typing import (
    Callable,
    ClassVar,
    Literal,
    Optional,
    Union)
from pydantic import BaseModel, Field


# ----------------------------------------------------------------------
# Constants
# ----------------------------------------------------------------------

SupportedCaptureTypes = Literal[
    "bool",
    "float",
    "int",
    "str",
    "ml-string",
    "list[float]",
    "list[int]",
    "list[bool]",
    "list[str]",
    "list[expand]",
    "expand"
    ]

SupportedChoiceTypes = Union[int, float, str]

# ----------------------------------------------------------------------
# Structures - User-facing models (external data)
# ----------------------------------------------------------------------

class BlueprintQuestion(BaseModel):
    """Used by an experiment builder (interview, etc).

    This should be part of a list of questions. Each BlueprintQuestion
    represents a single attribute that needs to be captured for
    experiment definition creation. 

    attribute: The associated attribute in __init__
    question: The question to ask the user
    entry_type: Helps with formatting the experience
    helper: A secondary string to aid the interview
    group: The group to expand. this is triggered if type is expand or
           list[expand]
    choices: restricts to a list of choices. ex: string must be 'cat' or
             'dog' or 3, 5, or 7
    min_value: if entry_type is numeric restricts the minimum value
    max_value: if entry_type is numeric restricts the maximum value
    """
    attribute:str
    question: str
    entry_type:SupportedCaptureTypes
    helper: Optional[str] = ""
    group: Optional[str] = None
    min_value: Optional[float] = None
    max_value: Optional[float] = None
    choices: list[SupportedChoiceTypes] = Field(default_factory=list)    
    default_value: Optional[SupportedChoiceTypes] = None


class FactoryOrder(BaseModel):
    """An experiment will define an 'order' for the factory. The ideal
    approach is to use the factory_id. But if this is not available then
    the order should have both a group and a name. This will then allow
    for loading dynamic components in the factory such as functions."""
    factory_id:str
    init_args: Optional[dict] = None


class FactoryEntry(BaseModel):
    """The entry within the component factory."""
    creator: Callable
    group: str
    tags: list[str] = Field(default_factory=list)
    summary: str
    factory_id:str = secrets.token_hex(4)
    interview_blueprint: list[BlueprintQuestion] = Field(default_factory=list)
    factory_can_init: bool = True

# ----------------------------------------------------------------------
# Structures - Internal helpers (non-validated)
# ----------------------------------------------------------------------


# ----------------------------------------------------------------------
# Protocols / Abstract Classes
# ----------------------------------------------------------------------

class SupportsFactory(ABC):
    """The foundation for any class based item. By inheriting this class
    in your code you are reducing the amount of code you need to write
    to register your custom Class. Of course, you can register without
    using this but it increases the amount of attributes you need to set
    so this is more a convinience than a requirement.
    """
    factory_id: ClassVar[str]
    group: ClassVar[str]
    tags: ClassVar[list[str]]  = []
    interview_blueprint: ClassVar[list[BlueprintQuestion]] = []
    summary: ClassVar[str] = "No summary found"
    factory_can_init: ClassVar[bool] = True



"""Built-in plotting utilities for ADGTK."""


from typing import Sequence
import matplotlib.pyplot as plt


def plot_single_line(
    data: Sequence,
    filename: str,
    data_label: str = "",
    title: str = "",
    x_label: str = "",
    y_label: str = "",
    show_legend: bool = False
) -> bool:
    """Plots a single line. basic plotting only

    :param data: the data to plot on th Y axis
    :type data: Sequence
    :param filename: the filename to save the image to
    :type filename: str
    :param data_label: if using a legend what is the data label,
        defaults to ""
    :type data_label: str, optional
    :param title: Title for the plot, defaults to ""
    :type title: str, optional
    :param x_label: label for the X-axis, defaults to ""
    :type x_label: str, optional
    :param y_label: label for the Y-axis, defaults to ""
    :type y_label: str, optional
    :param show_legend: show the legend?, defaults to False
    :type show_legend: bool, optional
    :return: True if plot attempted.
    :rtype: bool
    """

    x = range(len(data))
    if len(x) == 0:
        return False

    plt.plot(x, data, label=data_label)
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(title)
    if show_legend:
        plt.legend()
    plt.savefig(filename)
    return True



"""Processing functions for impairment of data."""

import random
from typing import Union

# ----------------------------------------------------------------------
# Constants
# ----------------------------------------------------------------------
EMPTY_PLACEHOLDER = "EMPTY_RECORD"
REPLACE_EMPTY_STRING_IN_DICT_WITH_PLACEHOLDER = True
RANDOM_CAN_SELECT_EMPTY = False

# ----------------------------------------------------------------------
# Helper Functions
# ----------------------------------------------------------------------

def get_random_key(
    data:dict,
    can_select_empty:bool=False,
    empty_placeholder:str="EMPTY_RECORD"
) -> Union[str, None]:
    """Returns a random key based on a dict. If it cannot then it will
    return None.

    :param data: The source data
    :type data: dict
    :param can_select_empty: ok to select an empty?, defaults to False
    :type can_select_empty: bool, optional
    :param empty_placeholder: empty data string, defaults to "EMPTY_RECORD"
    :type empty_placeholder: str, optional
    :return: A random key from the dict
    :rtype: str|None    
    """
    
    keys = list(data.keys())
    random.shuffle(keys)
    if can_select_empty:
        return random.choice(keys)

    # additional protections needed. Can return None now.
    for key in keys:
        if data[key] != empty_placeholder:
            return key
    # there is not a key that meets the criteria
    return None


def already_has_dropped(
    data:dict,
    empty_placeholder:str="EMPTY_RECORD"
) -> bool:
    """Checks to see if the dict already has a key is empty. True if the
    values contain the empty_placeholder.

    :param data: The source data
    :type data: dict
    :param empty_placeholder: empty data string, defaults to "EMPTY_RECORD"
    :type empty_placeholder: str, optional
    :return: T: that the dict has a key thats value is empty
    :rtype: bool
    """
    
    for _, value in data.items():
        if value == empty_placeholder:
            return True

    return False

def get_longest_string(
    data:dict,
    empty_placeholder:str="EMPTY_RECORD"
) -> Union[str, None]:
    """This function compares the different values and returns the key
    of the longest value.

    :param data: The source data
    :type data: dict
    :return: Key of the longest string
    :rtype: str|None
    """
    
    longest = None
    length = 0

    for key, item in data.items():
        if isinstance(item, str):
            if len(item) > length and item != empty_placeholder:
                longest = key
                length = len(item)

    return longest

# ----------------------------------------------------------------------
# Impairment helpers
# ----------------------------------------------------------------------


def drop_random_value(
    data:dict,
    can_select_empty:bool=False,
    empty_placeholder:str="EMPTY_RECORD"
    ) -> dict:
    """Impairs a dict by dropping a random value from the dict. If the
    data already has a dropped value, then it will not drop another.

    :param data: The source data
    :type data: dict
    :param can_select_empty: ok to select an empty?, defaults to False
    :type can_select_empty: bool, optional
    :param empty_placeholder: empty data string, defaults to "EMPTY_RECORD"
    :type empty_placeholder: str, optional
    :return: The updated dict
    :rtype: dict
    """
    
    if already_has_dropped(data=data, empty_placeholder=empty_placeholder):
        return data

    key = get_random_key(
        data=data, 
        can_select_empty=can_select_empty,
        empty_placeholder=empty_placeholder)
    
    if key is not None:
        data[key] = empty_placeholder        

    return data



"""Common Utils"""

from .cli import get_user_input, get_more_ask, create_line, clear_screen
from .file import load_data_from_csv_file
from .logs import (
    clear_agent_logs,
    clear_llm_logs,
    create_llm_logger,
    create_logger,
    get_scenario_logger
)
from .metrics import get_all_metric_data, get_single_run_metric_data


"""Defaults.

Should all be dict objects that can be used to generate configuration
files in the requested format YAML, TOML. etc.
"""
project_defaults = {
    "default_file_format": "toml",
    "experiment": {
        "data_dir": "data",
        "results_dir": "results",
        "tensorboard_dir": "runs",
        "experiment_definition_dir": "experiment_definitions"
    },
    "logging": {
        "log_dir": "logs",
        "level": "basic"
    }
}



"""Provides an Engine for measuring a dataset.

This module defines the `MeasurementEngine` class, which facilitates the 
measurement of data using various measurement types and factories. It also 
includes helper functions to validate measurement types.

TODO: Get the scenario logger and set it as the logger for the engine!
"""

from collections.abc import Iterable
import inspect
from logging import getLogger
from typing import (
    Any,
    Literal,
    Optional,
    TypedDict,
    Union,
    cast,
    get_args,
    get_origin
)
import uuid
import numpy as np
from adgtk.common.defaults import SCENARIO_LOGGER_NAME
from adgtk.common import UnableToMeasureException
from adgtk.tracking import ExperimentRunFolders, MetricTracker
from .factory import create_measurement
from .factory import (
    ClassBasedComparison,
    ClassBasedMeasurement,    
    MeasFactoryEntry,
    direct_comparison,    
    direct_measurement,
    distribution_measurement,
    distribution_comparison,
    get_measurements_by_tag,
    get_measurements_by_type,
    get_measurement_factory_entry,
    measurement_type,
    supports_factory
)

# ----------------------------------------------------------------------
# Development support only
# ----------------------------------------------------------------------
DEBUG = False

# ----------------------------------------------------------------------
# Structure
# ----------------------------------------------------------------------
calculation_type = Literal["avg", "sum", "max", "min", "raw", "distribution"]

class MeasurementData(TypedDict):
    """Records and reports on a single label/measurement and all its data"""
    label: str
    description: str
    data: list

class MeasurementReport(TypedDict):
    """A report structure that can be used by an agent to undertstand the
    measurements and associated results"""
    engine_id: str
    measurements: list[MeasurementData]


# ----------------------------------------------------------------------
# Helpers
# ----------------------------------------------------------------------
def supports_measurement_type(func, *args) -> bool:
    """
    Validates if the provided arguments match the expected types of a function.

    Args:
        func: The function whose parameter types are to be validated.
        *args: The arguments to validate against the function's parameter types.

    Returns:
        bool: True if all arguments match the expected types, False otherwise.
    """
    sig = inspect.signature(func)
    parameters = sig.parameters
    for i, (_, param) in enumerate(parameters.items()):
        if param.annotation != inspect.Parameter.empty:
            expected_type = param.annotation
            if i < len(args):
                arg = args[i]
                if DEBUG:
                    # Show the expected type and the argument type
                    print(f"Expected type: {expected_type}, "
                          f"Argument type: {type(arg)}")
                # Handle Union types
                if get_origin(expected_type) is Union:
                    if not any(isinstance(arg, t) for t in get_args(expected_type)):
                        return False
                # Handle regular types
                elif not isinstance(arg, expected_type):
                    return False

    if len(args) != len(parameters):
        return False
    
    return True
# ----------------------------------------------------------------------
# Engine
# ----------------------------------------------------------------------


class MeasurementEngine:
    """Drives measurements of data.

    The `MeasurementEngine` class manages the registration of measurement 
    factories, tracks metrics, and performs measurements on datasets.

    Attributes:
        engine_id (Optional[str]): A unique identifier for the engine.
        measurements (dict[str, supports_factory]): A dictionary of registered 
            measurement factories.
        details (dict[str, MeasFactoryEntry]): A dictionary containing details 
            about each registered measurement factory.
        metric_tracker (MetricTracker): Tracks metrics for the measurements.
        logger: Logger instance for logging engine-related events.
    """

    def __init__(
        self,
        engine_id: Optional[str] = None,
        add_factory_ids: Optional[list[str]] = None,
        add_by_type: Optional[measurement_type] = None,
        add_by_tag: Optional[Union[str, list[measurement_type]]] = None
    ) -> None:
        """
        Initializes the MeasurementEngine.

        Args:
            engine_id (Optional[str]): A unique identifier for the engine.
            add_factory_ids (Optional[list[str]]): A list of factory IDs to 
                register during initialization.
            add_by_type (Optional[measurement_type]): A measurement type to 
                register factories by.
            add_by_tag (Optional[Union[str, list[measurement_type]]]): A tag 
                or list of tags to register factories by.
        """
        self.engine_id = engine_id or str(uuid.uuid4())
        self.measurements: dict[str, supports_factory] = {}
        self.details: dict[str, MeasFactoryEntry] = {}
        self.metric_tracker = MetricTracker(
            name=self.engine_id,
            purpose="measurement"
        )
        self.logger = getLogger(SCENARIO_LOGGER_NAME)

        if add_factory_ids is not None:
            for entry in add_factory_ids:
                self.add(entry)
        if add_by_type is not None:
            entries = get_measurements_by_type(add_by_type)
            for entry in entries:
                self.add(entry['factory_id'])
        if add_by_tag is not None:
            entries = get_measurements_by_tag(add_by_tag)
            for entry in entries:                
                self.add(entry['factory_id'])

    def clear_results(self) -> None:
        """Clears all prior measurements but retains the definitions.
        """
        self.metric_tracker.clear_results()

    def get_all_data(self, label:str) -> list:
        """Gets all the data stored for a label

        Args:
            label (str): The label of the measurement

        Returns:
            list: the data
        """
        return self.metric_tracker.get_all_data(label)
    
    def measurement_count(self, label:str)-> int:
        """The number of measurements stored

        Args:
            label (str): the label of the measurement

        Returns:
            int: The count of data stored
        """
        return self.metric_tracker.measurement_count(label)

    def get_average(self, label:str) -> float:
        """Returns the average value for a measurement

        Args:
            label (str): the label of the measurement

        Returns:
            float: The average value
        """
        return self.metric_tracker.get_average(label)

    def get_latest_value(self, label:str)-> float:
        """Gets the latest value recorded

        Args:
            label (str): the label of the measurement

        Returns:
            float: The latest value
        """
        return self.metric_tracker.get_latest_value(label)

    def get_latest_distribution(self, label:str) -> np.ndarray:
        """Gets the latest distribution stored with a label.

        Args:
            label (str): the label of the measurement

        Returns:
            np.ndarray: The latest distribution
        """
        return self.metric_tracker.get_latest_distribution(label)

    def get_description(self, factory_id:str) -> str:
        """Retrieves the description of a measurement.

        Args:
            factory_id (str): The measurement to retrieve

        Returns:
            str: The description
        """
        if factory_id not in self.details.keys():
            raise IndexError("Unable to locate id %s", factory_id)
        
        entry = self.details[factory_id]
        return entry["description"]

        
    def add(self, factory_id: str, **kwargs) -> None:
        """
        Registers a measurement factory with the engine.

        Args:
            factory_id (str): The ID of the factory to register.
            **kwargs: Additional arguments for the factory creation.

        Raises:
            IndexError: If the factory ID is invalid or not found.
        """
        try:
            entry = get_measurement_factory_entry(factory_id)
            self.measurements[factory_id] = create_measurement(factory_id)
            self.details[factory_id] = entry
            self.metric_tracker.register_metric(label=factory_id)
        except IndexError:
            self.logger.error(
                "Measurement engine %s unable to add %s due to already "
                "registered. Ignoring request",
                self.engine_id, factory_id)

    def _update_tracker(
        self,
        label: str,
        results: list,
        record_as: calculation_type = "avg"
    ) -> None:
        """
        Records the results of a measurement in the tracker.

        Args:
            label (str): The tracker label.
            results (list): The data to add.
            record_as (calculation_type, optional): How to record the results. 
                Defaults to "avg".
        """
        if len(results) == 0:
            results = [0]
            self.logger.warning(
                "Measurement engine %s performed zero measurements.",
                self.engine_id)
        if record_as == "avg":
            value = sum(results) / len(results)
            self.metric_tracker.add_data(label=label, value=value)
        elif record_as == "max":
            value = max(results)
            self.metric_tracker.add_data(label=label, value=value)
        elif record_as == "min":
            value = min(results)
            self.metric_tracker.add_data(label=label, value=value)
        elif record_as == "sum":
            value = sum(results)
            self.metric_tracker.add_data(label=label, value=value)
        elif record_as == "raw":
            self.metric_tracker.add_raw_data(label=label, values=results)      

    def measure(
        self,
        data: Iterable,
        record_as: calculation_type = "avg"
    ) -> None:
        """
        Performs measurements on the provided dataset.

        Args:
            data (Iterable): The dataset to measure.
            record_as (calculation_type, optional): How to record the results. 
                Defaults to "avg".
        """        
        for label, meas in self.measurements.items():
            all_results = []
            if inspect.isclass(meas):
                meas = cast(ClassBasedMeasurement, meas)
            else:
                meas = cast(direct_measurement, meas)     

            # first, does the measurement want all the data?
            if supports_measurement_type(meas, data):
                try:
                    result = meas(data)
                    all_results.append(result)
                except UnableToMeasureException:
                    # NO-OP
                    pass
            else:           
                # if not, then iterate over the values
                # this is a fallback. measurements should be designed to
                # consider iterable values.
                for entry in data:
                    # Verify if the measurement type is supported
                    try:
                        if supports_measurement_type(meas, entry):                    
                            result = meas(entry)
                            if isinstance(result, (int, float)):
                                all_results.append(result)
                            elif isinstance(result, list):
                                all_results.extend(result)
                        else:
                            self.logger.warning(
                                f"{self.engine_id} No valid data for {label}. "
                                "skipping measure")
                            break
                    except UnableToMeasureException:
                        # NO-OP
                        pass
            self._update_tracker(
                label=label, results=all_results, record_as=record_as)

    def measure_dataset_distribution(self, dataset: Iterable) -> None:
        for label, meas in self.measurements.items():
            all_results = []
            if inspect.isclass(meas):
                meas = cast(ClassBasedMeasurement, meas)
            else:
                meas = cast(distribution_measurement, meas)
            # now measure
            if supports_measurement_type(meas, dataset):
                try:
                    result = meas(dataset)
                    all_results.append(result)
                except UnableToMeasureException:
                    # NO-OP
                    pass            
            self._update_tracker(
                label=label,
                results=all_results,
                record_as="distribution"
            )

    def compare_dataset_distribution(
        self,
        dataset_one: Iterable,
        dataset_two: Iterable,
        record_as: calculation_type = "avg"
    ) -> None:
        for label, meas in self.measurements.items():
            all_results = []
            if inspect.isclass(meas):
                meas = cast(ClassBasedComparison, meas)
            else:
                meas = cast(distribution_comparison, meas)
            if supports_measurement_type(meas, dataset_one, dataset_two):
                try:
                    result = meas(dataset_one, dataset_two)
                    all_results.append(result)
                except UnableToMeasureException:
                    # NO-OP
                    pass            
            self._update_tracker(
                label=label,
                results=all_results,
                record_as=record_as
            )
            
    def compare(
        self,
        data: Iterable[tuple[Any, Any]],
        record_as: calculation_type = "avg"
    ) -> None:
        """
        Performs measurements on the provided dataset.

        Args:
            data (Iterable): The dataset to measure.
            record_as (calculation_type, optional): How to record the results. 
                Defaults to "avg".
        """
        for label, meas in self.measurements.items():
            all_results = []
            if inspect.isclass(meas):
                meas = cast(ClassBasedComparison, meas)
            else:
                meas = cast(direct_comparison, meas)                
            for a,b in data:
                # Verify if the measurement type is supported
                if supports_measurement_type(meas, a, b):                    
                    result = meas(a, b)
                    if isinstance(result, (int, float)):
                        all_results.append(result)
                    elif isinstance(result, list):
                        all_results.extend(result)
            self._update_tracker(
                label=label, results=all_results, record_as=record_as)
            
    def save_data(self, folders: ExperimentRunFolders) -> None:
        """Saves the data to disk using the pre-defined folder structure

        Args:
            folders (ExperimentRunFolders): The results folders
        """
        self.metric_tracker.save_data(folders)

    def debug_report(self) -> None:
        """Prints a report to screen. Primary purpose is development.        
        """
        print(f"------- Measurement Engine: {self.engine_id} -------")
        max_key_length = max(len(k) for k in self.measurements.keys())
        for k, v in self.measurements.items():
            print(f"{k:<{max_key_length}}  {type(v).__name__}")

    def report(self) -> MeasurementReport:
        """Generates a measurement report based on the defined measurements
        and their associated data.
        
        Returns:
            MeasurementReport: A structured report containing engine ID and 
                all measurement data with labels, descriptions, and values.
        """
        measurements_data = []
        
        for factory_id in self.measurements.keys():
            # Get all data for this measurement
            data = self.get_all_data(factory_id)
            
            # Get description from details
            description = self.get_description(factory_id)
            
            measurement_data = MeasurementData(
                label=factory_id,
                description=description,
                data=data
            )
            measurements_data.append(measurement_data)
        
        return MeasurementReport(
            engine_id=self.engine_id,
            measurements=measurements_data
        )        

    def export_last_val_to_dict(self) -> dict:
        """Obtains the latest measurement as a dictionary.

        Returns:
            dict: the last value recorded by metric label
        """
        return self.metric_tracker.export_last_val_to_dict()



"""Provides and ADGTK managed dataset inventory.

Status: MVP, just enough to get basic data parsing now.

ROADMAP
=======
1. loading of datasets from HuggingFace ds, other wrappers.
"""

import argparse
import os
import sys
# before importing others
# ----------------------------------------------------------------------
# Start of path verification
# ----------------------------------------------------------------------
path = os.getcwd()
bootstrap_file = os.path.join(path, "bootstrap.py")
if not os.path.exists(bootstrap_file):
    print("ERROR: Unable to locate the bootstrap.py. Please check your path.")
    sys.exit(1)
# ----------------------------------------------------------------------
# End of path verification
# ----------------------------------------------------------------------
print("Here for some reason?")
from typing import Literal, Optional, Union, cast
from adgtk.data.structure import (
    SUPPORTED_FILE_ENCODING_TYPES,
    FileEncodingTypes
)


from adgtk.data.dataset import DatasetManager
from adgtk.utils import get_user_input
from adgtk.utils import create_logger


_logger = create_logger(
    logfile="file.manager.log",
    subdir="common",
    logger_name=__name__
)




# ----------------------------------------------------------------------
# Constants - user facing strings
# ----------------------------------------------------------------------

INTRO = ("ADGTK-DS: Bringing a consistent interface for your datasets\n"
         "===========================================================\n"
         "The adgtk-ds is a utility script for managing the datasets \n"
         "used for experimentation. This provides the human researcher the \n"
         "ability to refer to a file id during experiments instead of \ndealing"
         " with the complexity of the different formats, ingestion, \n"
         "etc as part of normal data processing.\n"
         )

HELP_W_INTRO = f"""{INTRO}\n
To use this script:\n
$ adgtk-ds [command] [options]

command
-------
- register : registers a new file into the manager
- retire   : retires a file from the manager
- report   : generates a report to the screen

For register the following are valid arguments:
--file         : the file to register, if not passed the user is asked
--encoding     : the encoding of the file,, if not passed the user is asked
                  (valid: {SUPPORTED_FILE_ENCODING_TYPES})
--metadata     : The metadata file w/path
                  (example: file.csv has file.meta.json)
--use          : The use of the dataset. 
                  (valid: test, train, validate, other)

for retire the following are valid arguments:
--id : the id of the file to retire

For find the following are valid arguments:
--filename (required) : the filename to register (w/path)

So for example, to register a file to the school file manager:
$ python file_manager.py school register --f=data/demo.csv -e=csv

report
------
you can further filter a report by passing in tags. For example:
$ python file_manager.py report ground entity

- Filters the report to the entries that have both tags ground and entity
- If no tags are sent all files are reported

Logging
-------
The log is maintained at logs/common/dataset.manager.log
"""


# ----------------------------------------------------------------------
# helper functions
# ----------------------------------------------------------------------


def parse_args() -> argparse.Namespace:
    """Parses the command line arguments

    Returns:
        argparse.Namespace: The command line input
    """
    parser = argparse.ArgumentParser()
    command_parser = parser.add_subparsers(
        dest="command", help="Command to perform")
    
    # Report subcommand
    report_parser = command_parser.add_parser(
        "report", help="Generate a report")
    report_parser.add_argument(
        "tags",
        nargs="*",
        help="Tag(s) to filter the report"
    )

    # Register subcommand
    register_parser = command_parser.add_parser(
        "register", help="Register a file")
    register_parser.add_argument(
        '-f', '--file',
        type=str,
        required=False,
        help="The full path to the file")
    register_parser.add_argument(
        '-e', '--encoding',
        type=str,
        required=False,
        choices=SUPPORTED_FILE_ENCODING_TYPES,
        help=f"The encoding of the file: {SUPPORTED_FILE_ENCODING_TYPES}")
    register_parser.add_argument(
        '-m', '--metadata',
        type=str,
        required=False,
        help="The full path to the metadata file")
    register_parser.add_argument(
        '--id',
        type=str,
        required=False,
        help="Optional custom ID to assign to the file entry"
    )
    register_parser.add_argument(
        "--use",
        type=str,
        choices=["test", "train", "validate", "other"],
        required=True,
        help="The use of the dataset",
        default="other"
    )

    # Retire subcommand
    retire_parser = command_parser.add_parser(
        "retire", help="Retire a file")
    retire_parser.add_argument(
        '--id', type=str, required=False, help="The file id")

    # Find subcommand
    find_parser = command_parser.add_parser("find", help="Find a file ID")
    find_parser.add_argument(
        '--filename', type=str, required=True, help="The filename")

    args = parser.parse_args()
    return args


def retire_file(file_tracker: DatasetManager, id: Optional[str] = None):

    if id is None:
        id = str(get_user_input(
            user_prompt="What is the file ID you would like to retire?",
            requested="str",
            helper="This is the internal file_id for the entry",
            allow_whitespace=False,
            min_characters=1
        ))
    try:
        file_tracker.retire_file(id)
    except IndexError as e:
        print(e)
        _logger.error(e)
        sys.exit()
    except (FileNotFoundError, PermissionError, IsADirectoryError, OSError) as e:
        print(e)
        _logger.error(e)
        print(f"Check logs/file.manager.log for additional information")
        sys.exit()

    print("File has been retired. File has been moved to .trash if "
          "stored internally")


def register_file(
    file_tracker: DatasetManager,
    source_file: Optional[str] = None,
    folder: Optional[str] = None,
    encoding: Optional[FileEncodingTypes] = None,
    metadata: Optional[bool] = None,
    tags: Optional[Union[str, list[str]]] = None,
    id: Optional[str] = None,
    use:Literal["test", "train", "validate", "other"] = "train"
):

    if source_file is None:
        source_file = str(get_user_input(
            user_prompt="Full path to the file to add",
            requested="str",
            helper="This is the full path to the file",
            allow_whitespace=False,
            min_characters=1
        ))
    file_w_path = source_file
    if not os.path.exists(file_w_path):
        print(f"File {file_w_path} not found")
        sys.exit(0)
    folder, filename = os.path.split(file_w_path)

    if encoding is None:
        encoding_str = str(get_user_input(
            user_prompt="Encoding",
            requested="str",
            choices=SUPPORTED_FILE_ENCODING_TYPES,
            helper="What encoding for the file."
        ))
        if encoding_str in SUPPORTED_FILE_ENCODING_TYPES:
            encoding = cast(FileEncodingTypes, encoding_str)
        else:
            raise ValueError("Failed to capture acceptable encoding")

    m_filename = None

    if metadata is None:
        # ask the question
        metadata = False
        meta_exists_input = str(get_user_input(
            user_prompt="Does a metadata file exist",
            requested="str",
            choices=["yes", "no"],
            helper="If no metadata file exists enter no, else yes"
        ))
        if meta_exists_input.lower() == "yes":
            metadata = True

    m_file_w_path = None
    if metadata:
        # There should be a metadata file
        file_ends = [".csv", ".pkl", ".json", ".txt", ".meta.json"]
        m_filename = filename
        for extension in file_ends:
            if filename.endswith(extension):
                m_filename = filename.removesuffix(extension)
        m_filename += ".meta.json"

        m_file_w_path = os.path.join(folder, m_filename)
        if not os.path.exists(m_file_w_path):
            print(f"WARNING: unable to find {m_file_w_path}. reverting to "
                  "NO for metadata file.")
            m_filename = None

    if id is not None and len(id) > 1:
        id = file_tracker.register(
            source_file=file_w_path,
            encoding=encoding,
            metadata_file=m_file_w_path,
            tags=tags,
            id=id,
            use=use
        )
    else:
        id = file_tracker.register(
            source_file=file_w_path,
            encoding=encoding,
            metadata_file=m_file_w_path,
            tags=tags,
            use=use
        )
    msg = f"Registered file as {id}"
    print(msg)
    _logger.info(msg)


def main():
    args = parse_args()

    # ensure the main folder is there
    os.makedirs(name=".tracking", exist_ok=True)

    ds_mgr = DatasetManager(folder=".tracking")

    if args.command == "report":
        ds_mgr.report(tag=args.tags)
    elif args.command == "register":
        register_file(
            file_tracker=ds_mgr,
            source_file=args.file,
            encoding=args.encoding,
            metadata=args.metadata,
            id=args.id,
            use=args.use
        )
    elif args.command == "find":
        if args.file is None:
            print(f"Filename is required to find an ID")
            sys.exit(1)

        id = ds_mgr.get_file_id(filename=args.file)
        print(f"ID for file is : {id}")
        sys.exit(0)

    elif args.command == "retire":
        if args.id is None:
            print(f"Filename is required to retire an ID")
            sys.exit(1)

        ds_mgr.retire_file(args.id)
    else:
        print(HELP_W_INTRO)



"""Utilities to ease record processing. 
"""

import logging
from typing import Literal, Any
from typing import (
    Literal,
    Final,
    Any,
    Sequence)
from adgtk.components.data import (
    RecordStore,
    CanFindRandomRecord,
    CanGetAllRecords,
    PresentableRecord,
    PresentableGroup)
from adgtk.instrumentation.base import (
    RecordAs,
    MeasInputType,
    SupportsMeasDef,
    InvalidMeasurementConfiguration)

# ----------------------------------------------------------------------
# Module Constants
# ----------------------------------------------------------------------

DEFAULT_MAX_IN_GROUP: Final[int] = 1000000  # for min/max checks if None


def convert_record_as(
    record_as: Literal["sum", "avg", "latest", "raw"]
) -> RecordAs:
    """Converts the string into an Enum

    :param record_as: the text mapping to RecordAs
    :type record_as: Literal[&quot;sum&quot;, &quot;avg&quot;, &quot;latest&quot;, &quot;raw&quot;]
    :raises InvalidMeasurementConfiguration: Unknown record_as value
    :return: The converted value
    :rtype: RecordAs
    """
    if record_as == "sum":
        return RecordAs.SUM
    elif record_as == "avg":
        return RecordAs.AVG
    elif record_as == "latest":
        return RecordAs.LATEST
    elif record_as == "raw":
        return RecordAs.RAW
    else:
        msg = f"Unknown record_as: {record_as} setting."
        raise InvalidMeasurementConfiguration(msg)


def process_as(data: Any) -> tuple:
    """Seeks to provide data

    :param data: The data to inspect
    :type data: Any
    :return: a tuple (process data, if iterable the items as)
    :rtype: Tuple[MeasInputType]
    """
    if isinstance(data, PresentableGroup):
        # A presentable group by design only has presentable records
        return (
            MeasInputType.PRESENTABLE_GROUP,
            MeasInputType.PRESENTABLE_RECORD)
    elif isinstance(data, PresentableRecord):
        return (MeasInputType.PRESENTABLE_RECORD,
                MeasInputType.NOOP)
    elif isinstance(data, str):
        return (MeasInputType.STRING, MeasInputType.NOOP)
    elif isinstance(data, dict):
        return (MeasInputType.DICTIONARY, MeasInputType.NOOP)
    elif isinstance(data, RecordStore):
        a = MeasInputType.UNKNOWN
        if len(data) == 0:
            return (MeasInputType.EMPTY, None)
        # get a sample
        elif isinstance(data, CanFindRandomRecord):
            random_record = data.find_random_record()
        elif isinstance(data, CanGetAllRecords):
            # lot slower, but better than unable to match
            random_record = data.get_all_records()
        else:
            logging.error("Unable to determine protocol for process_as")

        a, _ = process_as(random_record[0])
        return (MeasInputType.RECORD_STORE, a)

    elif isinstance(data, list):
        if len(data) == 0:
            return (MeasInputType.EMPTY, MeasInputType.NOOP)

        a, _ = process_as(data[0])

        if len(data) == 2:
            return (MeasInputType.PAIR_AS_LIST, a)

        return (MeasInputType.LIST, a)

    elif isinstance(data, Sequence):
        if len(data) == 0:
            return (MeasInputType.EMPTY, MeasInputType.NOOP)
        sample = data[0]
        a, _ = process_as(sample)
        if a is not MeasInputType.UNKNOWN:
            return (MeasInputType.ITERABLE, a)

        return (MeasInputType.ITERABLE, MeasInputType.NOOP)

    return (MeasInputType.UNKNOWN, MeasInputType.UNKNOWN)


def min_max_group_check(meas: SupportsMeasDef, data: PresentableGroup) -> bool:
    """checking for the director the different combinations makes the
    method is_measurable_with a bit too long. breaking out this code
    into a function for readability. This function validates whether
    the data and measurement requirements are met.

    :param meas: The measurement to check
    :type meas: SupportsMeasDef
    :type data: PresentableGroup
    :return: True if valid measurement for this group, else False
    :rtype: bool
    """
    min_val = meas.features.count_min
    max_val = meas.features.count_max

    if min_val is None:
        min_val = 0
    if max_val is None:
        max_val = DEFAULT_MAX_IN_GROUP

    record_count = len(data)

    if min_val <= record_count <= max_val:
        return True

    return False



"""Common structure. This module is used to define the common structures
that are used throughout the package.
"""

import logging
from typing import Protocol
from pydantic import BaseModel

class ResultsFile(BaseModel):
    filename: str
    purpose: str

# Do I need a metrics?

class ScenarioResult(BaseModel):
    files_written: list[ResultsFile]



class Scenario(Protocol):
    """The Scenario is the root of any experiment"""


    def start_scenario(self) -> ScenarioResult: ...
